# basis/_pod.py
"""Dimensionality reduction with Proper Orthogonal Decomposition (POD)."""

__all__ = [
    "PODBasis",
    "method_of_snapshots",
    "pod_basis",
    "svdval_decay",
    "cumulative_energy",
    "residual_energy",
]

import types
import warnings
import numpy as np
import scipy.linalg as la
import scipy.sparse as sparse
import sklearn.utils.extmath as sklmath
import matplotlib.pyplot as plt

from .. import errors, utils
from ._base import BasisTemplate
from ._linear import LinearBasis


# Helper functions ============================================================
requires_svdvals = utils.requires2(
    "svdvals",
    "no singular value data, call fit()",
)


def _Wmult(W, arr):
    """Matrix multiply ``W`` and ``arr``, where ``W`` may be a one-dimensional
    array representing diagonals or a two-dimensional array.
    """
    if W.ndim == 1:
        if arr.ndim == 1:
            return W * arr
        elif arr.ndim == 2:
            return W.reshape((-1, 1)) * arr
        else:
            raise ValueError("expected one- or two-dimensional array")
    return W @ arr


def method_of_snapshots(
    states,
    inner_product_matrix=None,
    minthresh: float = 1e-15,
    **options,
):
    r"""Use the method of snapshots to compute the left singular values of a
    collection of state snapshots.

    For a snapshot matrix :math:`\Q\in\RR^{n\times k}` (usually with
    :math:`n \ge k`) and a weighting matrix :math:`\W\in\RR^{n\times n}`,
    the method of snapshots computes the symmetric eigendecomposition

    .. math::
       \Q\trp\W\Q = \bfPsi\bfLambda\bfPsi\trp.

    The matrix :math:`\bfPsi\in\RR^{k\times k}` consists of the right singular
    vectors of :math:`\Q` and :math:`\bfLambda\in\RR^{k\times k}` is a diagonal
    matrix containing the square of the singular values of :math:`\Q`. The
    (weighted) left singular vectors are then given by
    :math:`\bfPhi = \Q\bfPsi\bfLambda^{-1/2} \in \RR^{n \times k}` and satisfy
    :math:`\Q = \bfPhi\bfLambda^{1/2}\bfPsi\trp` and
    :math:`\bfPhi\trp\W\bfPhi = \I`.

    Parameters
    ----------
    states : (n, k) ndarray,
        Snapshot matrix :math:`\Q` from which to compute the POD vectors.
    inner_product_matrix : (n, n) sparse SPD matrix or None
        Spatial inner product matrix :math:`\W` for measuring how different
        indices in the snapshot matrix interact with each other.
        If not provided, default to the standard Euclidean inner product
        (:math:`\W = \I`).
    minthresh : float > 0
        Threshold at which to truncate small eigenvalues. Singular vectors
        corresponding to eigenvalues that are less than this threshold are
        not included in the returned arrays.
    options : dict
        Additional arguments for :func:`scipy.linalg.eigh`.

    Returns
    -------
    V : (n, k') ndarray, k' <= k
        Left singular vectors :math:`\bfPhi`.
    svals : (k',) ndarray, k' <= k
        Singular values :math:`\operatorname{diag}(\bfLambda^{1/2})` in
        descending order.
    eigvecsT : (k', k') ndarray
        Transposed right singular vectors :math:`\bfPsi\trp`.

    Notes
    -----
    If, due to numerical precision errors, :func:`scipy.linalg.eigh` returns
    any negative eigenvalues, then ``minthresh`` is increased to the absolute
    value of the most negative eigenvalue.
    """
    n_states = states.shape[1]
    if inner_product_matrix is None:
        gramian = states.T @ (states / n_states)
    else:
        gramian = states.T @ _Wmult(inner_product_matrix, states / n_states)

    # Compute eigenvalue decomposition, using that the Gramian is symmetric.
    eigvals, eigvecs = la.eigh(gramian, **options)

    # Re-order (largest to smallest).
    eigvals = eigvals[::-1]
    eigvecs = eigvecs[:, ::-1]

    # NOTE: By definition the Gramian is symmetric positive semi-definite.
    # If any eigenvalues are smaller than zero, they are only measuring
    # numerical error and can be truncated.
    positives = eigvals > max(minthresh, abs(np.min(eigvals)))
    eigvecs = eigvecs[:, positives]
    eigvals = eigvals[positives]

    # Rescale and square root eigenvalues to get singular values.
    svals = np.sqrt(eigvals * n_states)
    V = states @ (eigvecs / svals)

    return V, svals, eigvecs.T


# Main class ==================================================================
class PODBasis(LinearBasis):
    r"""Proper othogonal decomposition basis, consisting of the principal left
    singular vectors of a collection of states.

    .. math::
       \text{svd}(\Q) = \bfPhi\bfSigma\bfPsi\trp
       \qquad\Longrightarrow\qquad
       \text{pod}(\Q, r) = \bfPhi_{:,:r}

    The low-dimensional approximation is linear, see :class:`LinearBasis`.
    Here, :math:`\Q\in\mathbb{R}^{n\times k}` is a collection of states,
    the only argument of :meth:`fit`.

    The POD basis entries matrix :math:`\Vr = \bfPhi_{:,:r}\in\RR^{n\times r}`
    always has orthonormal columns, i.e., :math:`\Vr\trp\Vr = \I`. If a weight
    matrix :math:`\W` is specified, a weighted SVD is computed so that
    :math:`\Vr\trp\W\Vr = \I`. The columns of the basis entries are also the
    dominant eigenvectors of :math:`\Q\trp\W\Q` and can be computed through
    eigendecomposition by setting ``svdsolver="eigh"``.

    The number of left singular vectors :math:`r` is the dimension of the
    reduced state and is set by specifying exactly one of the constructor
    arguments ``num_vectors``, ``svdval_threshold``, ``residual_energy``,
    ``cumulative_energy``, or ``projection_error``. Once the basis entries are
    set by calling :meth:`fit`, the reduced state dimension :math:`r` can be
    updated by calling :meth:`set_dimension`.

    The POD singular values, which are used to select :math:`r`, are the
    diagonals of :math:`\bfSigma` and are denoted
    :math:`\sigma_1,\ldots,\sigma_k`.

    Parameters
    ----------
    num_vectors : int
        Set the reduced state dimension :math:`r` to ``num_vectors``.
    svdval_threshold : float
        Choose :math:`r` as the number of normalized POD singular values that
        are greater than the given threshold, i.e.,
        :math:`\sigma_{i}/\sigma_{1} \ge` ``svdval_threshold`` for
        :math:`i=1,\ldots,r`.
    cumulative_energy : float
        Choose :math:`r` as the smallest integer such that
        :math:`\sum_{i=1}^{r}\sigma_i^2\big/\sum_{j=1}^{k}\sigma_j^2`
        is greater than or equal to ``cumulative_energy``.
    residual_energy : float
        Choose :math:`r` as the smallest integer such that
        :math:`\sum_{i=r+1}^k\sigma_i^2\big/\sum_{j=1}^k\sigma_j^2`
        is less than or equal to ``residual_energy``.
    projection_error : float
        Choose :math:`r` as the smallest integer such that
        :math:`\|\Q - \Vr\Vr\trp\Q\|_F \big/ \|\Q\|_F`
        is less than or equal to ``projection_error``.
    max_vectors : int
        Maximum number of POD basis vectors to store. After calling
        :meth:`fit`, the ``reduced_state_dimension`` can be increased up to
        ``max_vectors``. If not given (default), record all :math:`k` left
        singular vectors.
    svdsolver : str or callable
        Strategy for computing the thin SVD of the states.

        **Options:**

        * ``"dense"`` (default): Use :func:`scipy.linalg.svd` to
          compute the SVD. May be inefficient for very large state matrices.
        * ``"randomized"``: Compute an approximate SVD with a randomized
          approach via :func:`sklearn.utils.extmath.randomized_svd`.
          May be more efficient but less accurate for very large state
          matrices. **NOTE**: it is highly recommended to set ``max_vectors``
          to limit the number of computed singular vectors. In this case,
          only ``max_vectors`` singular *values* are computed as well, meaning
          the cumulative and residual energies cannot be computed exactly.
        * ``"method-of-snapshots"`` or ``"eigh"``: Compute the basis through a
          symmetric eigenvalue decomposition, rather than through the SVD, via
          :func:`scipy.linalg.eigh`. This is how POD was computed when it was
          orginally introduced. If the state dimension is larger than the
          number of snapshots, this method is much more efficient than the SVD.
          Moreover, non-Euclidean inner products (see :attr:`weights`)
          are handled much more efficiently this way than with an SVD-based
          approach. **NOTE**: in this case, an additional keyword argument
          ``minthresh`` defines a threshold at which small eigenvalues are
          truncated.
        * callable: If this argument is a callable function, use it for the
          SVD computation. The signature must match :func:`scipy.linalg.svd`,
          i.e., ``U, s, Vh = svdsolver(states, **svdsolver_options)``
    weights : (n, n) ndarray or (n,) ndarray None
        Weight matrix :math:`\W` or its diagonals.
        When provided, a weighted singular value decomposition of the states
        is used to ensure that the left singular vectors are orthogonal with
        respect to the weight matrix, i.e., :math:`\bfPhi\trp\W\bfPhi = \I`.
        If ``None`` (default), set :math:`\W` to the identity.
    name : str
        Label for the state variable that this basis approximates.
    svdsolver_options : dict
        Options to pass to the SVD solver.
    """

    # Valid SVD solvers.
    __SVDSOLVERS = types.MappingProxyType(
        {
            "dense": la.svd,
            "randomized": sklmath.randomized_svd,
            "method-of-snapshots": method_of_snapshots,
            # "streaming":  # TODO
        }
    )

    # Constructors ------------------------------------------------------------
    def __init__(
        self,
        num_vectors: int = None,
        svdval_threshold: float = None,
        cumulative_energy: float = None,
        residual_energy: float = None,
        projection_error: float = None,
        max_vectors: int = None,
        svdsolver: str = "dense",
        weights: np.ndarray = None,
        name: str = None,
        **svdsolver_options,
    ):
        """Initialize an empty basis."""
        # Superclass constructor.
        LinearBasis.__init__(self, entries=None, weights=weights, name=name)

        # Store dimension selection criteria.
        self._set_dimension_selection_criterion(
            num_vectors=num_vectors,
            svdval_threshold=svdval_threshold,
            cumulative_energy=cumulative_energy,
            residual_energy=residual_energy,
            projection_error=projection_error,
        )
        self.__energy_is_being_estimated = False

        # Initialize hyperparameter properties.
        if max_vectors is not None:
            max_vectors = int(max_vectors)
            if max_vectors <= 0:
                raise ValueError("max_vectors must be a positive integer")
        self.__max_vectors_desired = max_vectors
        self.svdsolver = svdsolver
        self.svdsolver_options = svdsolver_options

        # Initialize entry properties.
        self.__leftvecs = None
        self.__svdvals = None
        self.__rightvecs = None
        self.__cumulative_energy = None
        self.__residual_energy = None

        # Store weights (separate from LinearBasis.__weights)
        if (
            weights is not None
            and self.__svdsolverlabel != "method-of-snapshots"
        ):
            if weights.ndim == 1:
                self.__sqrt_weights = np.sqrt(weights)
            else:  # (weights.ndim == 2, checked by LinearBasis)
                self.__sqrt_weights = la.sqrtm(weights)
                self.__sqrt_weights_cho = la.cho_factor(self.__sqrt_weights)

    @classmethod
    def from_svd(
        cls,
        leftvecs: np.ndarray,
        svdvals: np.ndarray,
        rightvecs: np.ndarray = None,
        num_vectors: int = None,
        svdval_threshold: float = None,
        residual_energy: float = None,
        cumulative_energy: float = None,
        projection_error: float = None,
        max_vectors: int = None,
        weights: np.ndarray = None,
        name: str = None,
    ):
        r"""Initialize a :class:`PODBasis` from the singular value
        decomposition of an :math:`n\times k` matrix.

        Parameters
        ----------
        leftvecs : (n, r) ndarray
            Left singular vectors.
        svdvals : (r,) ndarray
            Singular values.
        rightvecs : (k, r) ndarray or None
            Right singular vectors. Each *column* is a singular vector.

        Returns
        -------
        Initialized :class:`PODBasis` object.

        Notes
        -----
        See :class:`PODBasis` for details on other arguments.
        """
        # Default dimensionality criterion: use all left singular vectors.
        if all(
            arg is None
            for arg in (
                num_vectors,
                svdval_threshold,
                residual_energy,
                cumulative_energy,
                projection_error,
            )
        ):
            num_vectors = leftvecs.shape[1]

        basis = cls(
            num_vectors=num_vectors,
            svdval_threshold=svdval_threshold,
            residual_energy=residual_energy,
            cumulative_energy=cumulative_energy,
            projection_error=projection_error,
            max_vectors=max_vectors,
            weights=weights,
            name=name,
        )
        basis._store_svd(leftvecs, svdvals, rightvecs)
        basis._set_dimension_from_criterion()

        return basis

    # Properties: hyperparameters ---------------------------------------------
    @property
    def svdsolver(self) -> str:
        """Strategy for computing the thin SVD of the states, either
        ``'dense'``, ``'randomized'``, or ``'custom'``.
        """
        return self.__svdsolverlabel

    @svdsolver.setter
    def svdsolver(self, s):
        if callable(s):
            self.__svdsolverlabel = "custom"
            self.__svdengine = s
            return

        if s == "eigh":
            s = "method-of-snapshots"

        if s not in self.__SVDSOLVERS:
            raise AttributeError(
                f"invalid svdsolver '{s}', options: "
                + ", ".join(self.__SVDSOLVERS.keys())
            )
        self.__svdsolverlabel = s
        self.__svdengine = self.__SVDSOLVERS[s]

    @property
    def svdsolver_options(self) -> dict:
        """Options to pass to the SVD solver."""
        return self.__svdsolver_options

    @svdsolver_options.setter
    def svdsolver_options(self, options):
        if options is None:
            options = dict()
        if not isinstance(options, dict):
            raise TypeError("svdsolver_options must be a dictionary")
        self.__svdsolver_options = options

    # Properties: entries -----------------------------------------------------
    @property
    def leftvecs(self):
        """Leading left singular vectors of the training data."""
        return self.__leftvecs

    @property
    def max_vectors(self) -> int:
        """Number of POD basis vectors stored in memory.
        The ``reduced_state_dimension`` may be increased up to ``max_vectors``.
        """
        return None if self.leftvecs is None else self.leftvecs.shape[1]

    @property
    def svdvals(self):
        """Normalized singular values of the training data."""
        return self.__svdvals

    @property
    def rightvecs(self):
        """Leading *right* singular vectors of the training data,
        if available.
        """
        return self.__rightvecs

    @property
    def cumulative_energy(self) -> float:
        r"""Amount of singular value energy captured by the basis,
        :math:`\sum_{i=1}^r\sigma_i^2\big/\sum_{j=1}^k\sigma_j^2`.
        """
        return self.__cumulative_energy

    @property
    def residual_energy(self) -> float:
        r"""Amount of singular value energy *not* captured by the basis,
        :math:`\sum_{i=r+1}^k\sigma_i^2\big/\sum_{j=1}^k\sigma_j^2`.
        """
        return self.__residual_energy

    def __str__(self):
        """String representation: class, dimensions, and
        singular value energies.
        """
        out = [LinearBasis.__str__(self)]

        if (ce := self.cumulative_energy) is not None:
            if self.__energy_is_being_estimated:
                out.append(f"approximate cumulative energy: {ce:%}")
            else:
                out.append(f"cumulative energy: {ce:%}")

        if (re := self.residual_energy) is not None:
            if self.__energy_is_being_estimated:
                out.append(f"approximate residual energy:   {re:.4e}")
            else:
                out.append(f"residual energy:   {re:.4e}")

        if (mv := self.max_vectors) is not None:
            out.append(f"{mv:d} basis vectors available")

        if self.__svdsolverlabel == "dense":
            out.append("SVD solver: scipy.linalg.svd()")
        elif self.__svdsolverlabel == "randomized":
            out.append("SVD solver: sklearn.utils.extmath.randomized_svd()")
        elif self.__svdsolverlabel == "custom":
            if (name := self.__svdengine.__name__) == "<lambda>":
                out.append("SVD solver: custom lambda function")
            else:
                out.append(f"SVD solver: {name}()")

        return "\n  ".join(out)

    # Dimension selection -----------------------------------------------------
    def _set_dimension_selection_criterion(
        self,
        num_vectors: int = None,
        svdval_threshold: float = None,
        cumulative_energy: float = None,
        residual_energy: float = None,
        projection_error: float = None,
    ):
        args = [
            ("num_vectors", num_vectors),
            ("svdval_threshold", svdval_threshold),
            ("cumulative_energy", cumulative_energy),
            ("residual_energy", residual_energy),
            ("projection_error", projection_error),
        ]
        provided = [(arg[1] is not None) for arg in args]

        # More than one argument provided.
        if sum(provided) > 1:
            firstarg = args[np.argmax(provided)]
            warnings.warn(
                "received multiple dimension selection criteria, using "
                f"{firstarg[0]}={firstarg[1]}",
                errors.OpInfWarning,
            )
            self.__criterion = firstarg
            return

        # Return the one provided argument.
        for name, val in args:
            if val is not None:
                self.__criterion = (name, val)
                return

        # No arguments provided.
        raise ValueError(
            "exactly one dimension selection criterion must be provided"
        )

    @BasisTemplate.reduced_state_dimension.setter
    def reduced_state_dimension(self, r):
        """Set the reduced state dimension :math:`r`. If there is SVD data,
        set the basis entries to the first r left singular vectors.

        If r > max_vectors, set r = max_vectors and raise a warning.
        """
        r = int(r)

        # No basis data yet, but when fit() is called there will be r vectors.
        if self.svdvals is None:
            self._set_dimension_selection_criterion(num_vectors=r)
            BasisTemplate.reduced_state_dimension.fset(self, r)
            return

        # Basis data already exists, change the dimension and update.
        if r > (k := self.max_vectors):
            warnings.warn(
                "selected reduced dimension exceeds number of stored vectors, "
                f"setting reduced_state_dimension = max_vectors = {k:d}",
                errors.OpInfWarning,
            )
            r = self.max_vectors
        BasisTemplate.reduced_state_dimension.fset(self, r)

        # Update singular value energies.
        r = self.reduced_state_dimension
        svdvals2 = self.svdvals**2
        self.__cumulative_energy = np.sum(svdvals2[:r]) / np.sum(svdvals2)
        self.__residual_energy = 1 - self.__cumulative_energy

        # Update entries.
        LinearBasis.__init__(
            self,
            entries=self.__leftvecs[:, :r],
            weights=self.weights,
            check_orthogonality=True,
            name=self.name,
        )

    def _set_dimension_from_criterion(self):
        """Set the dimension by interpreting the ``__criterion`` attribute."""
        criterion, value = self.__criterion
        svdvals2 = self.svdvals**2
        nsvdvals = svdvals2.size
        energy = np.cumsum(svdvals2) / np.sum(svdvals2)

        if criterion == "num_vectors":
            r = value
        elif criterion == "svdval_threshold":
            r = np.count_nonzero(self.svdvals >= value)
        elif criterion == "cumulative_energy":
            r = int(np.searchsorted(energy, value)) + 1
            if self.__energy_is_being_estimated:
                warnings.warn(
                    "cumulative energy is being estimated from only "
                    f"{nsvdvals:d} singular values",
                    errors.OpInfWarning,
                )
        elif criterion == "residual_energy":
            r = np.count_nonzero(1 - energy >= value) + 1
            if self.__energy_is_being_estimated:
                warnings.warn(
                    "residual energy is being estimated from only "
                    f"{nsvdvals:d} singular values",
                    errors.OpInfWarning,
                )
        elif criterion == "projection_error":
            r = np.count_nonzero(np.sqrt(1 - energy) >= value) + 1
            if self.__energy_is_being_estimated:
                warnings.warn(
                    "projection error is being estimated from only "
                    f"{nsvdvals:d} singular values",
                    errors.OpInfWarning,
                )

        self.reduced_state_dimension = r

    def set_dimension(
        self,
        num_vectors: int = None,
        svdval_threshold: float = None,
        cumulative_energy: float = None,
        residual_energy: float = None,
        projection_error: float = None,
    ):
        r"""Set the reduced state dimension :math:`r`.
        Exactly one argument should be specified

        Parameters
        ----------
        num_vectors : int
            Set :math:`r` to ``num_vectors``.
        svdval_threshold : float
            Choose :math:`r` as the number of normalized POD singular values
            that are greater than the given threshold, i.e.,
            :math:`\sigma_{i}/\sigma_{1} \ge` ``svdval_threshold`` for
            :math:`i=1,\ldots,r`.
        cumulative_energy : float
            Choose :math:`r` as the smallest integer such that
            :math:`\sum_{i=1}^{r}\sigma_i^2\big/\sum_{j=1}^{k}\sigma_j^2`
            is greater than or equal to ``cumulative_energy``.
        residual_energy : float
            Choose :math:`r` as the smallest integer such that
            :math:`\sum_{i=r+1}^k\sigma_i^2\big/\sum_{j=1}^k\sigma_j^2`
            is less than or equal to ``residual_energy``.
        projection_error : float
            Choose :math:`r` as the smallest integer such that
            :math:`\|\Q - \Vr\Vr\trp\Q\|_F \big/ \|\Q\|_F`
            is less than or equal to ``projection_error``, where :math:`\Q`
            is the matrix of training snapshots.
        """
        self._set_dimension_selection_criterion(
            num_vectors=num_vectors,
            svdval_threshold=svdval_threshold,
            residual_energy=residual_energy,
            cumulative_energy=cumulative_energy,
            projection_error=projection_error,
        )

        # No basis data yet, do nothing.
        if self.svdvals is None:
            return

        # Basis data exists, set the dimension and update.
        self._set_dimension_from_criterion()

    # Fitting -----------------------------------------------------------------
    def _store_svd(self, left, svdvals, right):
        """Store the singular value decomposition, normalizing the singular
        values.
        """
        self.__leftvecs = left
        svdvals = np.sort(svdvals)[::-1]
        self.__svdvals = svdvals / svdvals[0]
        self.__rightvecs = right

    def fit(self, states):
        """Compute the POD basis entries by taking the SVD of the states.

        Parameters
        ----------
        states : (n, k) ndarray
            Matrix of :math:`k` :math:`n`-dimensional snapshots.
        """
        if np.ndim(states) == 3:
            states = np.hstack(states)

        # Limit maximum number of vectors if needed.
        rmax = min(states.shape)
        keep = self.__max_vectors_desired
        if keep is None:
            keep = rmax
        elif keep > rmax:
            warnings.warn(
                f"only {rmax:d} singular vectors can be extracted "
                f"from ({states.shape[0]:d} x {states.shape[1]:d}) snapshots, "
                f"setting max_vectors={rmax:d}",
                errors.OpInfWarning,
            )
            keep = rmax

        # SVD solver settings.
        self.__energy_is_being_estimated = False
        options = self.svdsolver_options.copy()

        if self.__svdsolverlabel == "dense":
            options["full_matrices"] = False
        elif self.__svdsolverlabel == "randomized":
            options["n_components"] = keep
            if "random_state" not in options:
                options["random_state"] = None
            if keep < rmax:
                self.__energy_is_being_estimated = True
        elif self.__svdsolverlabel == "method-of-snapshots":
            options["inner_product_matrix"] = self.weights

        # Weight the states.
        if (
            self.weights is not None
            and self.__svdsolverlabel != "method-of-snapshots"
        ):
            if states.shape[0] != (nW := self.__sqrt_weights.shape[0]):
                raise errors.DimensionalityError(
                    f"states not aligned with weights, should have {nW:d} rows"
                )
            states = _Wmult(self.__sqrt_weights, states)

        # Compute the SVD.
        V, svdvals, Wt = self.__svdengine(states, **options)

        # Unweight the basis.
        if (
            self.weights is not None
            and self.__svdsolverlabel != "method-of-snapshots"
        ):
            if self.__sqrt_weights.ndim == 1:
                V = _Wmult(1 / self.__sqrt_weights, V)
            else:
                V = la.cho_solve(self.__sqrt_weights_cho, V)

        # Store the results.
        self._store_svd(
            left=V[:, :keep],
            svdvals=svdvals,
            right=Wt[:keep, :].T,
        )
        self._set_dimension_from_criterion()

        return self

    # Visualization -----------------------------------------------------------
    @requires_svdvals
    def _plot_single(self, plotter, **kwargs):
        """Execute a single plotting routine."""
        plotter(self.svdvals, **kwargs)
        return ax if (ax := kwargs.get("ax")) is not None else plt.gca()

    def plot_svdval_decay(
        self,
        threshold=None,
        right: int = None,
        ax: plt.Axes = None,
        **options,
    ):
        """Plot the normalized singular value decay.

        Parameters
        ----------
        threshold : float or list[float] or None
            Cutoff value(s) to mark on the plot.
        right : int or None
            Maximum singular value index to plot (``plt.xlim(right=right)``).
        ax : matplotlib.Axes or None
            Axes to plot on.
            If ``None`` (default), a new single-axes figure is created.
        options : dict
            Options to pass to :func:`matplotlib.pyplot.semilogy`.

        Returns
        -------
        ax : matplotlib.Axes
            Axes for the plot.
        """
        kwargs = dict(threshold=threshold, plot=True, right=right, ax=ax)
        kwargs.update(options)
        return self._plot_single(svdval_decay, **kwargs)

    def plot_cumulative_energy(
        self,
        threshold=None,
        right: int = None,
        ax: plt.Axes = None,
        **options,
    ):
        r"""Plot the cumulative singular value energy and a function of the
        basis size.

        The cumulative energy of :math:`r` singular values is defined by

        .. math::
           \kappa_r = \sum_{i=1}^r\sigma_i^2 \bigg/ \sum_{j=1}^k\sigma_j^2.

        This method plots :math:`\kappa_r` as a function of :math:`r`.

        Parameters
        ----------
        threshold : float or list[float] or None
            Threshold energy value(s) to mark on the plot.
        right : int or None
            Maximum singular value index to plot (``plt.xlim(right=right)``).
        ax : matplotlib.Axes or None
            Axes to plot on.
            If ``None`` (default), a new single-axes figure is created.
        kwargs : dict
            Options to pass to :func:`matplotlib.pyplot.semilogy`.

        Returns
        -------
        ax : matplotlib.Axes
            Axes for the plot.
        """
        kwargs = dict(threshold=threshold, plot=True, right=right, ax=ax)
        kwargs.update(options)
        return self._plot_single(cumulative_energy, **kwargs)

    def plot_residual_energy(
        self,
        threshold=None,
        right: int = None,
        ax: plt.Axes = None,
        **options,
    ):
        r"""Plot the residual singular value energy as a function of the basis
        size.

        The residual energy of :math:`r` singular values is defined by

        .. math::
           \epsilon_r
           = \sum_{i=r+1}^k\sigma_i^2 \bigg/ \sum_{j=1}^k\sigma_j^2
           = 1 - \left(
           \sum_{i=1}^r\sigma_i^2 \bigg/ \sum_{j=1}^k\sigma_j^2
           \right)

        This method plots :math:`\epsilon_r` as a function of :math:`r`.

        Parameters
        ----------
        threshold : float or list[float] or None
            Cutoff value(s) to mark on the plot.
        right : int or None
            Maximum singular value index to plot (``plt.xlim(right=right)``).
        ax : matplotlib.Axes or None
            Axes to plot on.
            If ``None`` (default), a new single-axes figure is created.
        options : dict
            Options to pass to :func:`matplotlib.pyplot.semilogy`.

        Returns
        -------
        ax : matplotlib.Axes
            Axes for the plot.
        """
        kwargs = dict(threshold=threshold, plot=True, right=right, ax=ax)
        kwargs.update(options)
        return self._plot_single(residual_energy, **kwargs)

    def plot_projection_error(
        self,
        threshold=None,
        right: int = None,
        ax: plt.Axes = None,
        **options,
    ):
        r"""Plot the relative projection error of the training snapshots as a
        function of the basis size.

        The relative projection error for the rank-:math:`r` POD basis
        corresponding to the training snapshot matrix
        :math:`\Q\in\RR^{n\times k}` is defined by

        .. math::
           \rho_r = \frac{\|\Q - \Vr\Vr\trp\Q\|_{F}}{\|\Q\|_{F}},

        where :math:`\Vr\in\RR^{n \times r}` are the basis entries.
        This method plots :math:`\rho_r` as a function of :math:`r`;
        :math:`\rho_r` is calculated via the singular values:

        .. math::
           \rho_r
            = \sqrt{\frac{\sum_{j = r + 1}^{\ell}\sigma_{j}^{2}}{
            \sum_{j=1}^{\ell}\sigma_{j}^{2}}}

        Parameters
        ----------
        threshold : float or list[float] or None
            Cutoff value(s) to mark on the plot.
        right : int or None
            Maximum singular value index to plot (``plt.xlim(right=right)``).
        ax : matplotlib.Axes or None
            Axes to plot on.
            If ``None`` (default), a new single-axes figure is created.
        options : dict
            Options to pass to :func:`matplotlib.pyplot.semilogy`.

        Returns
        -------
        ax : matplotlib.Axes
            Axes for the plot.

        Notes
        -----
        This method shows the projection error of the training snapshots.
        See :meth:`projection_error` to calculate the projection error for an
        arbitrary snapshot or collection of snapshots.
        """
        kwargs = dict(
            threshold=threshold,
            plot=True,
            right=right,
            ax=ax,
            sqrt=True,
        )
        kwargs.update(options)
        return self._plot_single(residual_energy, **kwargs)

    @requires_svdvals
    def plot_energy(self, right: int = None):
        """Plot the normalized singular values and the cumulative and residual
        energies.

        Parameters
        ----------
        right : int or None
            Maximum singular value index to plot (``plt.xlim(right=right)``).
        """
        r = self.reduced_state_dimension

        def _rline(ax, ymin):
            ax.axvline(r, color="gray", linewidth=0.5, zorder=1)
            ax.text(
                r + 0.5,
                ymin,
                f"r = {r}",
                color="gray",
                horizontalalignment="left",
                verticalalignment="bottom",
            )

        fig, axes = plt.subplots(1, 2, figsize=(13.44, 4.8))

        ax = self.plot_svdval_decay(right=right, ax=axes[0])
        ax.set_title("POD singular values")
        _rline(ax, 1.05 * ax.get_ylim()[0])

        ax = self.plot_residual_energy(right=right, ax=axes[1])
        ax.set_ylabel("Residual energy", color="C1")
        ax.spines["right"].set_visible(True)
        ax.tick_params(axis="y", which="both", color="C1", labelcolor="C1")
        ax.set_title("POD singular value energy")

        ax = self.plot_cumulative_energy(
            right=right,
            ax=axes[1].twinx(),
            color="C0",
        )
        ax.spines["left"].set_visible(True)
        ax.set_ylabel("Cumulative energy", color="C0")
        ax.tick_params(axis="y", which="both", color="C0", labelcolor="C0")
        _rline(ax, 0.01 + ax.get_ylim()[0])

        fig.tight_layout()

    # Persistence -------------------------------------------------------------
    def save(self, savefile, overwrite=False):
        """Save the basis to an HDF5 file.

        Parameters
        ----------
        savefile : str
            Path of the file to save the basis to.
        overwrite : bool
            If ``True``, overwrite the file if it already exists.
            If ``False`` (default), raise a ``FileExistsError`` if the file
            already exists.
        """
        with utils.hdf5_savehandle(savefile, overwrite) as hf:
            meta = hf.create_dataset("meta", shape=(0,))
            if self.name:
                meta.attrs["name"] = self.name

            meta.attrs["criterion_type"] = self.__criterion[0]
            hf.create_dataset("criterion_value", data=[self.__criterion[1]])

            if self.__max_vectors_desired is not None:
                hf.create_dataset(
                    "max_vectors",
                    data=[self.__max_vectors_desired],
                )

            if (w := self.weights) is not None:
                if isinstance(w, sparse.dia_array):
                    w = w.data[0]
                hf.create_dataset("weights", data=w)

            if self.leftvecs is not None:
                hf.create_dataset("leftvecs", data=self.leftvecs)
                hf.create_dataset("svdvals", data=self.svdvals)
                hf.create_dataset("rightvecs", data=self.rightvecs)

    @classmethod
    def load(cls, loadfile, max_vectors=None):
        """Load a basis from an HDF5 file.

        Parameters
        ----------
        loadfile : str
            Path to the file where the basis was stored via :meth:`save`.
        max_vectors : int or None
            Maximum number of POD vectors to load.
            If ``None`` (default), load all stored vectors.

        Returns
        -------
        PODBasis object
        """
        kwargs = {}
        with utils.hdf5_loadhandle(loadfile) as hf:
            attrs = hf["meta"].attrs

            if "name" in attrs:
                kwargs["name"] = attrs["name"]

            kwargs[attrs["criterion_type"]] = hf["criterion_value"][0]

            if "max_vectors" in hf:
                kwargs["max_vectors"] = hf["max_vectors"][0]

            if "weights" in hf:
                kwargs["weights"] = hf["weights"][:]

            if "leftvecs" in hf:
                if (r := max_vectors) is not None:
                    r = min(r, hf["leftvecs"].shape[1])
                    if "num_vectors" in kwargs:
                        kwargs["num_vectors"] = r
                kwargs["leftvecs"] = hf["leftvecs"][:, :r]
                kwargs["svdvals"] = hf["svdvals"][:]
                kwargs["rightvecs"] = hf["rightvecs"][:, :r]
                return cls.from_svd(**kwargs)

            return cls(**kwargs)


# Functional API ==============================================================
def pod_basis(
    states,
    num_vectors: int = None,
    svdval_threshold: float = None,
    cumulative_energy: float = None,
    residual_energy: float = None,
    projection_error: float = None,
    svdsolver: str = "dense",
    weights: np.ndarray = None,
    return_rightvecs: bool = False,
    **svdsolver_options,
):
    r"""Compute a POD basis from the given states.

    Parameters
    ----------
    states : (n, k) ndarray
        Matrix of :math:`k` :math:`n`-dimensional snapshots.
    return_rightvec : bool
        If ``True``, return the right singular vectors as well.

    Returns
    -------
    basis : (n, r) ndarray
        POD basis matrix, the first :math:`r` left singular vectors of the
        states.
    svdvals : (n,), (k,), or (r,) ndarray
        Normalized singular values in descending order.
        Always returns as many as are calculated:
        :math:`\min\{n, k\}` for ``svdsolver="dense"``,
        :math:`r` for ``svdsolver="randomized"``.
    rightvecs : (k, r) ndarray
        First :math:`r` **right** singular vectors, as columns.
        **Only returned** if ``return_rightvecs=True``.

    Notes
    -----
    See :class:`PODBasis` for the mathematical definition of POD and for
    details on other function arguments.
    """
    # If no criteria given, use the maximum dimension based on the states.
    criteria = {
        num_vectors,
        svdval_threshold,
        cumulative_energy,
        residual_energy,
        projection_error,
    }
    if len(criteria) == 1 and criteria.pop() is None:
        num_vectors = min(states.shape)

    basis = PODBasis(
        num_vectors=num_vectors,
        svdval_threshold=svdval_threshold,
        cumulative_energy=cumulative_energy,
        residual_energy=residual_energy,
        projection_error=projection_error,
        svdsolver=svdsolver,
        weights=weights,
        **svdsolver_options,
    ).fit(states)

    if return_rightvecs:
        r = basis.reduced_state_dimension
        return basis.entries, basis.svdvals, basis.rightvecs[:, :r]

    return basis.entries, basis.svdvals


def svdval_decay(
    singular_values,
    threshold: float = 1e-8,
    plot: bool = True,
    right: int = None,
    ax: plt.Axes = None,
    **kwargs,
):
    """Count the number of **normalized** singular values that are greater than
    a specified threshold.

    Parameters
    ----------
    singular_values : (n,) ndarray
        Singular values of a snapshot matrix, e.g.,
        ``scipy.linalg.svdvals(states)``.
    threshold : float or list[float]
        Cutoff value(s) for the singular values.
    plot : bool
        If ``True``, plot the singular values and the cutoff value(s) against
        the singular value index.
    right : int or None
        Maximum singular value index to plot (``plt.xlim(right=right)``).
    ax : matplotlib.Axes or None
        Axes to plot the results on if ``plot=True``.
        If not given, a new single-axes figure is created.
    kwargs : dict
        Options to pass to :func:`matplotlib.pyplot.semilogy`.

    Returns
    -------
    ranks : int or list[int]
        Number of singular values greater than the cutoff value(s).
    """
    singular_values = np.sort(singular_values)[::-1]
    singular_values /= singular_values[0]

    # Calculate the number of singular values above the cutoff value(s).
    if threshold:
        if one_threshold := np.isscalar(threshold):
            threshold = [threshold]
        ranks = [np.count_nonzero(singular_values > ep) for ep in threshold]

    if plot:
        # Plot singular values.
        if ax is None:
            ax = plt.figure().add_subplot(111)

        options = dict(
            marker="*",
            color="k",
            markersize=8,
            markeredgewidth=0,
            zorder=3,
        )
        options.update(kwargs)
        marker = options.pop("marker")

        j = np.arange(1, singular_values.size + 1)
        ax.semilogy(j, singular_values, marker, **options)
        ax.set_xlim((0, j.size + 1))
        if right:
            right = min(right, j.size)
            ax.set_xlim(right=right)
            _idx = min(int(right), singular_values.size - 1)
            ax.set_ylim(bottom=singular_values[_idx] / 5)

        # Draw cutoff value(s).
        if threshold:
            rborder = ax.get_xlim()[1]
            ylim = ax.get_ylim()
            for sigma, r in zip(threshold, ranks):
                ax.axhline(sigma, color="gray", linewidth=0.5)
                ax.text(
                    rborder - 0.5,
                    1.05 * sigma,
                    f"{sigma:.2e}",
                    color="gray",
                    horizontalalignment="right",
                    verticalalignment="bottom",
                )
                ax.axvline(r, color="gray", linewidth=0.5)
                ax.text(
                    r + 0.5,
                    1.05 * ylim[0],
                    f"r = {r}",
                    color="gray",
                    horizontalalignment="left",
                    verticalalignment="bottom",
                )
            ax.set_ylim(ylim)

        ax.set_xlabel("Singular value index")
        ax.set_ylabel("Normalized singular values")

    if threshold:
        return ranks[0] if one_threshold else ranks


def cumulative_energy(
    singular_values,
    threshold: float = 0.9999,
    plot: bool = True,
    right: int = None,
    ax: plt.Axes = None,
    **kwargs,
):
    r"""Compute the number of singular values needed to surpass a given
    cumulative energy threshold.

    The cumulative energy of :math:`r` singular values is defined by

    .. math::
       \kappa_r = \sum_{i=1}^r\sigma_i^2 \bigg/ \sum_{j=1}^k\sigma_j^2.

    This function determines the smalled :math:`r` such that
    :math:`\kappa_r \ge` ``threshold``.

    Parameters
    ----------
    singular_values : (n,) ndarray
        Singular values of a snapshot matrix, e.g.,
        ``scipy.linalg.svdvals(states)``.
    threshold : float or list[float]
        Energy capture threshold(s). Default is 99.99%.
    plot : bool
        If ``True``, plot the cumulative energy and the capture threshold(s)
        against the singular value index (linear scale).
    right : int or None
        Maximum singular value index to plot (``plt.xlim(right=right)``).
    ax : matplotlib.Axes or None
        Axes to plot the results on if ``plot=True``.
        If not given, a new single-axes figure is created.
    kwargs : dict
        Options to pass to :func:`matplotlib.pyplot.plot`.

    Returns
    -------
    ranks : int or list[int]
        Number of singular values required to capture the specified energy.
    """
    # Calculate the cumulative energy.
    svdvals2 = np.sort(singular_values)[::-1] ** 2
    energy = np.cumsum(svdvals2) / np.sum(svdvals2)
    energy = np.concatenate(([0], energy))

    # Determine the points at which the cumulative energy passes the threshold.
    if threshold:
        if one_threshold := np.isscalar(threshold):
            threshold = [threshold]
        ranks = [int(np.searchsorted(energy, xi)) for xi in threshold]

    if plot:
        # Plot cumulative energy and threshold value(s).
        if ax is None:
            ax = plt.figure().add_subplot(111)

        options = dict(
            linestyle="-",
            marker="o",
            color="C2",
            markersize=4,
            markeredgewidth=0,
            linewidth=0.5,
            zorder=3,
        )
        options.update(kwargs)

        j = np.arange(singular_values.size + 1)
        ax.plot(j, energy, **options)
        ax.set_xlim(0, j.size)
        ylim = (ax.get_ylim()[0], 1.05)
        if right:
            right = min(right, j.size)
            ax.set_xlim(right=right)

        if threshold:
            rborder = ax.get_xlim()[1]
            for kappa, r in zip(threshold, ranks):
                ax.axhline(kappa, color="gray", linewidth=0.5)
                ax.text(
                    rborder - 0.5,
                    kappa + 0.01,
                    f"{kappa:%}",
                    color="gray",
                    horizontalalignment="right",
                    verticalalignment="bottom",
                )
                ax.axvline(r, color="gray", linewidth=0.5)
                ax.text(
                    r + 0.5,
                    ylim[0] + 0.01,
                    f"r = {r}",
                    color="gray",
                    horizontalalignment="left",
                    verticalalignment="bottom",
                )

        ax.set_ylim(ylim)
        ax.set_xlabel(r"Singular value index")
        ax.set_ylabel(r"Cumulative energy")

    if threshold:
        return ranks[0] if one_threshold else ranks


def residual_energy(
    singular_values,
    threshold: float = 1e-6,
    plot: bool = True,
    right: int = None,
    ax: plt.Axes = None,
    sqrt: bool = False,
    **kwargs,
):
    r"""Compute the number of singular values needed such that the residual
    energy drops beneath a given threshold.

    The residual energy of :math:`r` singular values is defined by

    .. math::
        \epsilon_r
        = \sum_{i=r+1}^k\sigma_i^2 \bigg/ \sum_{j=1}^k\sigma_j^2
        = 1 - \left(
        \sum_{i=1}^r\sigma_i^2 \bigg/ \sum_{j=1}^k\sigma_j^2
        \right)

    This function determines the smallest :math:`r` such that
    :math:`\epsilon_r \le` ``threshold``.

    Parameters
    ----------
    singular_values : (n,) ndarray
        Singular values of a snapshot matrix, e.g.,
        ``scipy.linalg.svdvals(states)``.
    threshold : float or list[float]
        Energy residual threshold(s). Default is 10^-6.
    plot : bool
        If ``True``, plot the residual energy and the threshold(s)
        against the singular value index.
    right : int or None
        Maximum singular value index to plot (``plt.xlim(right=right)``).
    ax : matplotlib.Axes or None
        Axes to plot the results on if ``plot=True``.
        If not given, a new single-axes figure is created.
    sqrt : bool
        If ``True``, square root the residual energies to get the projection
        error of the training snapshots.
    kwargs : dict
        Options to pass to :func:`matplotlib.pyplot.semilogy`.

    Returns
    -------
    ranks : int or list[int]
        Number of singular values required to shrink the residual energy below
        the specified threshold.
    """
    # Calculate the residual energy.
    svdvals2 = np.sort(singular_values)[::-1] ** 2
    res_energy = 1 - (np.cumsum(svdvals2) / np.sum(svdvals2))
    res_energy = np.concatenate(([1], res_energy))
    if sqrt:
        res_energy = np.sqrt(res_energy)

    # Determine the points when the residual energy dips under the threshold.
    if threshold:
        if one_threshold := np.isscalar(threshold):
            threshold = [threshold]
        ranks = [np.count_nonzero(res_energy > eps) for eps in threshold]

    if plot:
        # Plot residual energy and threshold value(s).
        if ax is None:
            ax = plt.figure().add_subplot(111)

        options = dict(
            linestyle="-",
            marker="s",
            color="C1",
            markersize=4,
            markeredgewidth=0,
            linewidth=0.5,
            zorder=3,
        )
        options.update(kwargs)

        j = np.arange(singular_values.size + 1)
        ax.semilogy(j, res_energy, **options)
        ax.set_xlim(0, j.size)
        bottom = res_energy[-2]
        if right:
            right = min(right, j.size)
            ax.set_xlim(right=right)
            bottom = res_energy[min(int(right) + 1, res_energy.size - 3)]
        bottom = max(bottom, 1e-17)
        ylim = (bottom, 1.1)

        if threshold:
            rborder = ax.get_xlim()[1]
            for epsilon, r in zip(threshold, ranks):
                ax.axhline(epsilon, color="gray", linewidth=0.5)
                ax.text(
                    rborder - 0.5,
                    1.1 * epsilon,
                    f"{epsilon:.2e}",
                    color="gray",
                    horizontalalignment="right",
                    verticalalignment="bottom",
                )
                ax.axvline(r, color="gray", linewidth=0.5)
                ax.text(
                    r + 0.5,
                    1.1 * ylim[0],
                    f"r = {r}",
                    color="gray",
                    horizontalalignment="left",
                    verticalalignment="bottom",
                )

        ax.set_ylim(ylim)
        ax.set_xlabel("Singular value index")
        if sqrt:
            ax.set_ylabel("Relative projection error")
        else:
            ax.set_ylabel("Residual energy")

    if threshold:
        return ranks[0] if one_threshold else ranks



# ddt/_interpolation.py
"""Time derivative estimators based on interpolation."""

__all__ = [
    "InterpDerivativeEstimator",
]


import types
import warnings
import numpy as np
import scipy.interpolate as interp

from .. import errors
from ._base import DerivativeEstimatorTemplate


class InterpDerivativeEstimator(DerivativeEstimatorTemplate):
    r"""Time derivative estimator based on interpolation.

    For a set of (compressed) snapshots
    :math:`\Qhat = [~\qhat_0~~\cdots~~\qhat_{k-1}~] \in \RR^{r \times k}`,
    this class forms one-dimensional differentiable interpolants for each of
    the :math:`r` rows of :math:`\Qhat`.

    Parameters
    ----------
    time_domain : (k,) ndarray
        Time domain of the snapshot data.
    InterpolatorClass : str or scipy.interpolate class
        Class for performing the interpolation. Must obey the following syntax:

            >>> intrp = InterpolatorClass(time_domain, states, **options)
            >>> new_states = intrp(new_time_domain, 0)
            >>> state_ddts = intrp(new_time_domain, 1)

        The following strings are also accepted.

        * ``"cubic"`` (default): use :class:`scipy.interpolate.CubicSpline`.
        * ``"akima"``: use :class:`scipy.interpolate.Akima1DInterpolator`.
          This is a local interpolation method and is more resitant to
          outliers than :class:`scipy.interpolate.CubicSpline`. However, it is
          not recommended if the time points are not uniformly spaced.
        * ``"pchip"``: use :class:`scipy.interpolate.PchipInterpolator`.
          The interpolator preserves monotonicity in the interpolation data
          and does not overshoot if the data is not smooth.
    new_time_domain : (k',) ndarray or None
        If given, evaluate the interpolator at these points to generate new
        state snapshots and corresponding time derivatives. If input snapshots
        are also given, interpolate them as well and evaluate the interpolant
        at these new points. The `new_time_domain` should lie within the range
        of the original `time_domain`; a warning is raised if extrapolation is
        requested.
    options : dict
        Keyword arguments for the constructor of the :attr:`InterpolatorClass`.
    """

    _interpolators = types.MappingProxyType(
        {
            "akima": interp.Akima1DInterpolator,
            "cubic": interp.CubicSpline,
            "pchip": interp.PchipInterpolator,
        }
    )

    # Constructor -------------------------------------------------------------
    def __init__(
        self,
        time_domain: np.ndarray,
        InterpolatorClass="cubic",
        new_time_domain: np.ndarray = None,
        **options,
    ):
        """Set the time domain, InterpolatorClass, and options."""
        DerivativeEstimatorTemplate.__init__(self, time_domain)

        # Set the InterpolatorClass.
        if not isinstance(InterpolatorClass, type):
            if InterpolatorClass not in self._interpolators:
                options = ", ".join([repr(k) for k in self._interpolators])
                raise ValueError(
                    f"invalid InterpolatorClass '{InterpolatorClass}', "
                    f"options: {options}"
                )
            InterpolatorClass = self._interpolators[InterpolatorClass]
            options["axis"] = 1
        self.__IC = InterpolatorClass

        # Set the new_time_domain.
        if (t2 := new_time_domain) is not None:
            if not isinstance(t2, np.ndarray) or t2.ndim != 1:
                raise TypeError(
                    "new_time_domain must be a one-dimensional array or None"
                )
            t1 = time_domain
            if t2.min() < t1.min() or t2.max() > t1.max():
                warnings.warn(
                    "new_time_domain extrapolates beyond time_domain",
                    errors.OpInfWarning,
                )
        self.__t2 = new_time_domain

        # Set the interpolator constructor options.
        self.__options = types.MappingProxyType(options)

    # Properties --------------------------------------------------------------
    def __str__(self):
        """String representation: class name, time domain."""
        head = DerivativeEstimatorTemplate.__str__(self)
        options = ", ".join(
            [
                f"{key}={repr(value)}"
                for key, value in self.options.items()
                if key != "axis"
            ]
        )
        tail = [
            f"InterpolatorClass: {self.InterpolatorClass.__name__}({options})"
        ]
        if (t2 := self.new_time_domain) is not None:
            tail.append(
                f"new_time_domain: {t2.size} entries "
                f"in [{t2.min()}, {t2.max()}]"
            )
        return f"{head}\n  " + "\n  ".join(tail)

    @property
    def InterpolatorClass(self):
        """One-dimensional differentiable interpolator class."""
        return self.__IC

    @property
    def new_time_domain(self):
        """Time domain at which to evaluate the interpolator and its
        first derivative.
        """
        return self.__t2

    @property
    def options(self):
        """Keyword arguments for the constructor of the
        :attr:`InterpolatorClass`.
        """
        return self.__options

    # Main routine ------------------------------------------------------------
    def estimate(self, states, inputs=None):
        """Estimate the first time derivatives of the states.

        Parameters
        ----------
        states : (r, k) ndarray
            State snapshots, either full or (preferably) reduced.
        inputs : (m, k) ndarray or None
            Inputs corresponding to the state snapshots, if applicable.

        Returns
        -------
        _states : (r, k') ndarray
            State snapshots or state estimates at the :attr:`new_time_domain`.
        ddts : (r, k') ndarray
            First time derivatives corresponding to ``_states``.
        _inputs : (m, k') ndarray or None
            Inputs corresponding to ``_states``, if applicable.
            **Only returned** if ``inputs`` is provided.
        """
        states, inputs = self._check_dimensions(states, inputs)

        statespline = self.InterpolatorClass(
            self.time_domain,
            states,
            **self.options,
        )

        t = self.time_domain
        if (t2 := self.new_time_domain) is not None:
            states = statespline(t2)
            if inputs is not None:
                inputspline = self.InterpolatorClass(
                    t,
                    inputs,
                    **self.options,
                )
                inputs = inputspline(t2)
            t = self.new_time_domain

        ddts = statespline(t, 1)

        if inputs is None:
            return states, ddts
        return states, ddts, inputs



# post/_errors.py
"""Tools for accuracy and error evaluation."""

__all__ = [
    "projection_error",
    "frobenius_error",
    "lp_error",
    "Lp_error",
]

import numpy as np
import scipy.linalg as la
import scipy.integrate as spintegrate


def _absolute_and_relative_error(Qtrue, Qapprox, norm):
    """Compute the absolute and relative errors between Qtrue and Qapprox,
    where Qapprox approximates Qtrue:

        absolute_error = ||Qtrue - Qapprox||,
        relative_error = ||Qtrue - Qapprox|| / ||Qtrue||
                       = absolute_error / ||Qtrue||,

    with ||Q|| defined by norm(Q).
    """
    norm_of_data = norm(Qtrue)
    absolute_error = norm(Qtrue - Qapprox)
    return absolute_error, absolute_error / norm_of_data


def projection_error(states, basis):
    """Calculate the absolute and relative projection errors induced by
    projecting states to a low dimensional basis, i.e.,

        absolute_error = ||Q - Vr Vr^T Q||_F,
        relative_error = ||Q - Vr Vr^T Q||_F / ||Q||_F

    where Q = states and Vr = basis. Note that Vr Vr^T is the orthogonal
    projector onto subspace of R^n defined by the basis.

    Parameters
    ----------
    states : (n, k) or (k,) ndarray
        Matrix of k snapshots where each column is a single snapshot, or a
        single 1D snapshot. If 2D, use the Frobenius norm; if 1D, the l2 norm.
    Vr : (n, r) ndarray
        Low-dimensional basis of rank r. Each column is one basis vector.

    Returns
    -------
    absolute_error : float
        Absolute projection error ||Q - Vr Vr^T Q||_F.
    relative_error : float
        Relative projection error ||Q - Vr Vr^T Q||_F / ||Q||_F.
    """
    Qapprox = basis @ (basis.T @ states)
    return _absolute_and_relative_error(states, Qapprox, la.norm)


def frobenius_error(Qtrue, Qapprox):
    """Compute the absolute and relative Frobenius-norm errors between the
    snapshot sets Qtrue and Qapprox, where Qapprox approximates Qtrue:

        absolute_error = ||Qtrue - Qapprox||_F,
        relative_error = ||Qtrue - Qapprox||_F / ||Qtrue||_F.

    Parameters
    ----------
    Qtrue : (n, k)
        "True" data. Each column is one snapshot, i.e., Qtrue[:, j] is the data
        at some time t[j].
    Qapprox : (n, k)
        An approximation to Qtrue, i.e., Qapprox[:, j] approximates Qtrue[:, j]
        and corresponds to some time t[j].

    Returns
    -------
    abs_err : float
        Absolute error ||Qtrue - Qapprox||_F.
    rel_err : float
        Relative error ||Qtrue - Qapprox||_F / ||Qtrue||_F.
    """
    # Check dimensions.
    if Qtrue.shape != Qapprox.shape:
        raise ValueError("Qtrue and Qapprox not aligned")
    if Qtrue.ndim != 2:
        raise ValueError("Qtrue and Qapprox must be two-dimensional")

    # Compute the errors.
    return _absolute_and_relative_error(
        Qtrue, Qapprox, lambda Z: la.norm(Z, ord="fro")
    )


def lp_error(Qtrue, Qapprox, p=2, normalize=False):
    """Compute the absolute and relative lp-norm errors between the snapshot
    sets Qtrue and Qapprox, where Qapprox approximates to Qtrue:

        absolute_error_j = ||Qtrue_j - Qapprox_j||_p,
        relative_error_j = ||Qtrue_j - Qapprox_j||_p / ||Qtrue_j||_p.

    Parameters
    ----------
    Qtrue : (n, k) or (n,) ndarray
        "True" data. Each column is one snapshot, i.e., Qtrue[:, j] is the data
        at some time t[j]. If one-dimensional, all of Qtrue is a single
        snapshot.
    Qapprox : (n, k) or (n,) ndarray
        An approximation to Qtrue, i.e., Qapprox[:, j] approximates Qtrue[:, j]
        and corresponds to some time t[j]. If one-dimensional, all of Qapprox
        is a single snapshot approximation.
    p : float
        Order of the lp norm (default p=2 is the Euclidean norm). Used as
        the `ord` argument for scipy.linalg.norm(); see options at
        docs.scipy.org/doc/scipy/reference/generated/scipy.linalg.norm.html.
    normalize : bool
        If true, compute the normalized absolute error instead of the relative
        error, defined by

            normalized_absolute_error_j
                = ||Qtrue_j - Qapprox_j||_2 / max_k{||Qtrue_k||_2}.

    Returns
    -------
    abs_err : (k,) ndarray or float
        Absolute error of each pair of snapshots Qtrue[:, j] and Qapprox[:, j].
        If Qtrue and Qapprox are one-dimensional, Qtrue and Qapprox are treated
        as single snapshots, so the error is a float.
    rel_err : (k,) ndarray or float
        Relative or normed absolute error of each pair of snapshots Qtrue[:, j]
        and Qapprox[:, j]. If Qtrue and Qapprox are one-dimensional, Qtrue and
        Qapprox are treated as single snapshots, so the error is a float.
    """
    # Check p.
    if not np.isscalar(p) or p <= 0:
        raise ValueError("norm order p must be positive (np.inf ok)")

    # Check dimensions.
    if Qtrue.shape != Qapprox.shape:
        raise ValueError("Qtrue and Qapprox not aligned")
    if Qtrue.ndim not in (1, 2):
        raise ValueError("Qtrue and Qapprox must be one- or two-dimensional")

    # Compute the error.
    norm_of_data = la.norm(Qtrue, ord=p, axis=0)
    if normalize:
        norm_of_data = norm_of_data.max()
    absolute_error = la.norm(Qtrue - Qapprox, ord=p, axis=0)
    return absolute_error, absolute_error / norm_of_data


def Lp_error(Qtrue, Qapprox, t=None, p=2):
    """Compute the absolute and relative Lp-norm error (with respect to time)
    between the snapshot sets Qtrue and Qapprox, where Qapprox approximates
    Qtrue:

        absolute_error = ||Qtrue - Qapprox||_{L^p},
        relative_error = ||Qtrue - Qapprox||_{L^p} / ||Qtrue||_{L^p},

    where

        ||Z||_{L^p} = (int_{t} ||z(t)||_{p} dt)^{1/p},          p < infinity,
        ||Z||_{L^p} = sup_{t}||z(t)||_{p},                      p = infinity.

    The trapezoidal rule is used to approximate the integrals (for finite p).
    This error measure is only consistent for data sets where each snapshot
    represents function values, i.e.,

        Qtrue[:, j] = [q(t1), q(t2), ..., q(tk)]^T.

    Parameters
    ----------
    Qtrue : (n, k) or (k,) ndarray
        "True" data corresponding to time t. Each column is one snapshot,
        i.e., Qtrue[:, j] is the data at time t[j]. If one-dimensional, each
        entry is one snapshot.
    Qapprox : (n, k) or (k,) ndarray
        An approximation to Qtrue, i.e., Qapprox[:, j] approximates Qtrue[:, j]
        and corresponds to time t[j]. If one-dimensional, each entry is one
        snapshot.
    t : (k,) ndarray
        Time domain of the data Qtrue and the Qapprox.
        Required unless p == np.inf.
    p : float > 0
        Order of the Lp norm. May be infinite (np.inf).

    Returns
    -------
    abs_err : float
        Absolute error ||Qtrue - Qapprox||_{L^p}.
    rel_err : float
        Relative error ||Qtrue - Qapprox||_{L^p} / ||Qtrue||_{L^p}.
    """
    # Check p.
    if not np.isscalar(p) or p <= 0:
        raise ValueError("norm order p must be positive (np.inf ok)")

    # Check dimensions.
    if Qtrue.shape != Qapprox.shape:
        raise ValueError("Qtrue and Qapprox not aligned")
    if Qtrue.ndim == 1:
        Qtrue = np.atleast_2d(Qtrue)
        Qapprox = np.atleast_2d(Qapprox)
    elif Qtrue.ndim > 2:
        raise ValueError("Qtrue and Qapprox must be one- or two-dimensional")

    # Pick the norm based on p.
    if 0 < p < np.inf:
        if t is None:
            raise ValueError("time t required for p < infinty")
        if t.ndim != 1:
            raise ValueError("time t must be one-dimensional")
        if Qtrue.shape[-1] != t.shape[0]:
            raise ValueError("Qtrue not aligned with time t")

        def pnorm(Z):
            Zpnorm = np.sum(np.abs(Z) ** p, axis=0)
            return spintegrate.trapezoid(Zpnorm, t) ** (1 / p)

    else:  # p == np.inf

        def pnorm(Z):
            return np.max(np.abs(Z), axis=0).max()

    # Compute the error.
    return _absolute_and_relative_error(Qtrue, Qapprox, pnorm)



# basis/_linear.py
"""Linear basis class."""

__all__ = [
    "LinearBasis",
]

import warnings
import numpy as np
import scipy.sparse as sparse
import matplotlib.pyplot as plt

from .. import errors, utils
from ._base import BasisTemplate


requires_entries = utils.requires2(
    "entries",
    "basis entries not initialized",
)


class LinearBasis(BasisTemplate):
    r"""Linear low-dimensional state approximation.

    This class approximates high-dimensional states :math:`\q\in\RR^n` as a
    linear combination of :math:`r` basis vectors
    :math:`\v_1,\ldots,\v_r\in\RR^n`. The basis matrix
    :math:`\Vr = [~\v_1~~\cdots~~\v_r~]\in \RR^{n \times r}` and an (optional)
    weighting matrix :math:`\W\in\RR^{n \times n}` define the approximation.

    The encoding from the high-dimensional space :math:`\RR^n` to the
    low-dimensional space :math:`\RR^r` is given by

    .. math::
       \q \mapsto \qhat = \Vr\trp\W\q,

    while the decoding from low-dimensional space :math:`\RR^r` to the
    high-dimensional space :math:`\RR^n` is defined as

    .. math::
       \qhat \mapsto \breve{\q} = \Vr\qhat = \sum_{i=1}^r \hat{q}_i \v_i,

    where :math:`\qhat = [\hat{q}_1,\ldots,\hat{q}_r]\trp\in\RR^r`.

    Basis entries :math:`\Vr` and the weights :math:`\W` are specified
    explicitly in the constructor, not learned from state data.

    Parameters
    ----------
    entries : (n, r) ndarray
        Basis entries :math:`\Vr\in\RR^{n\times r}`.
    weights : (n, n) ndarray or (n,) ndarray None
        Weight matrix :math:`\W\in\RR^n` or its diagonal entries.
        If ``None`` (default), set :math:`\W` to the identity.
    check_orthogonality : bool
        If ``True``, raise a warning if the basis is not orthogonal, i.e.,
        if :math:`\Vr\trp\W\Vr` is not the identity.
    name : str or None
        Label for the state variable that this basis approximates.

    Notes
    -----
    Pair with a :class:`opinf.pre.ShiftScaleTransformer` to do centered
    approximations of the form :math:`\q \approx\Vr\qhat + \bar{\q}`.
    """

    def __init__(
        self,
        entries,
        weights=None,
        check_orthogonality: bool = True,
        name: str = None,
    ):
        """Initialize the basis entries."""
        BasisTemplate.__init__(self, name=name)

        # Empty intializer for child classes (POD).
        if entries is None:
            self.__entries = None
            self.__weights = weights
            return

        # Set the entries.
        self.__entries = entries
        BasisTemplate.full_state_dimension.fset(self, entries.shape[0])
        BasisTemplate.reduced_state_dimension.fset(self, entries.shape[1])

        # Set the weights.
        if weights is not None:
            if (dim := np.ndim(weights)) == 1:
                n = np.size(weights)
                weights = sparse.dia_array(([weights], [0]), shape=(n, n))
            elif dim != 2:
                raise ValueError("expected one- or two-dimensional weights")
        self.__weights = weights

        # Verify orthogonality if desired.
        if check_orthogonality:
            V, W = self.entries, self.weights
            Id = (V.T @ V) if W is None else (V.T @ W @ V)
            if not np.allclose(Id, np.eye(self.reduced_state_dimension)):
                warnings.warn("basis not orthogonal", errors.OpInfWarning)

    # Properties --------------------------------------------------------------
    @property
    def entries(self):
        r"""Entries of the basis matrix :math:`\Vr\in\RR^{n \times r}`. Also
        accessible via indexing (``basis[:]``).
        """
        return self.__entries

    @property
    def weights(self) -> np.ndarray:
        r"""Weight matrix :math:`\W \in \RR^{n \times n}`."""
        return self.__weights

    @property
    def full_state_dimension(self):
        r"""Dimension :math:`n` of the full state."""
        return BasisTemplate.full_state_dimension.fget(self)

    @property
    def reduced_state_dimension(self):
        r"""Dimension :math:`r` of the reduced (compressed) state."""
        return BasisTemplate.reduced_state_dimension.fget(self)

    @requires_entries
    def __getitem__(self, key):
        """self[:] --> self.entries."""
        return self.entries[key]

    def fit(self, *args, **kwargs):
        """Do nothing, the basis entries are set in the constructor."""
        return self

    # Dimension reduction -----------------------------------------------------
    @requires_entries
    def compress(self, state: np.ndarray) -> np.ndarray:
        r"""Map high-dimensional states to low-dimensional latent coordinates.

        .. math:: \q \mapsto \qhat = \Vr\trp\q.

        If a weight matrix :math:`\W` is present, the compression is

        .. math:: \q \mapsto \qhat = \Vr\trp\W\q.

        Parameters
        ----------
        states : (n, ...) ndarray
            Matrix of `n`-dimensional state vectors, or a single state vector.

        Returns
        -------
        states_compressed : (r, ...) ndarray
            Matrix of `r`-dimensional latent coordinate vectors, or a single
            coordinate vector.
        """
        if self.weights is not None:
            state = self.weights @ state
        return self.entries.T @ state

    @requires_entries
    def decompress(
        self,
        states_compressed: np.ndarray,
        locs=None,
    ) -> np.ndarray:
        r"""Map low-dimensional latent coordinates to high-dimensional states.

        .. math:: \qhat \mapsto \breve{\q} = \Vr\qhat

        Parameters
        ----------
        states_compressed : (r, ...) ndarray
            Matrix of `r`-dimensional latent coordinate vectors, or a single
            coordinate vector.
        locs : slice or (p,) ndarray of integers or None
            If given, return the decompressed state at only the `p` specified
            locations (indices) described by ``locs``.

        Returns
        -------
        states_decompressed : (n, ...) or (p, ...) ndarray
            Matrix of `n`-dimensional decompressed state vectors, or the `p`
            entries of such at the entries specified by ``locs``.
        """
        Vr = self.entries if locs is None else self.entries[locs]
        return Vr @ states_compressed

    # Visualizations ----------------------------------------------------------
    @requires_entries
    def plot1D(self, x=None, num_vectors=None, ax=None, **kwargs):
        """Plot the basis vectors over a one-dimensional domain.

        Parameters
        ----------
        x : (n,) ndarray or None
            One-dimensional spatial domain over which to plot the vectors.
            Defaults to [0, 1] with `n` points.
        num_vectors : int or None
            Number of basis vectors to plot.
            If ``None`` (default), plot all basis vectors.
        ax : plt.Axes or None
            Matplotlib Axes to plot on.
            If ``None`` (default), a new figure is created.
        kwargs : dict
            Other keyword arguments to pass to ``plt.plot()``.

        Returns
        -------
        ax : plt.Axes
            Matplotlib Axes for the plot.
        """
        if x is None:
            x = np.linspace(0, 1, self.full_state_dimension)
        if num_vectors is None:
            num_vectors = self.reduced_state_dimension
        num_vectors = min(num_vectors, self.reduced_state_dimension)
        if ax is None:
            ax = plt.figure().add_subplot(111)

        for j in range(num_vectors):
            ax.plot(x, self.entries[:, j], **kwargs)
        ax.set_xlim(x[0], x[-1])
        ax.set_xlabel("spatial domain")
        ax.set_ylabel("basis vectors")

        return ax

    # Persistence -------------------------------------------------------------
    def __eq__(self, other) -> bool:
        """Two LinearBasis objects are equal if their type, dimensions, and
        basis entries are the same.
        """
        if not isinstance(other, self.__class__):
            return False
        if self.shape != other.shape:
            return False
        if self.weights is None and other.weights is not None:
            return False
        if (w1 := self.weights) is not None:
            if (w2 := other.weights) is None:
                return False
            if sparse.issparse(w1) and sparse.issparse(w2):
                w1, w2 = w1.data[0], w2.data[0]
            if not np.allclose(w1, w2):
                return False
        return np.all(self.entries == other.entries)

    def save(self, savefile: str, overwrite: bool = False):
        """Save the basis to an HDF5 file.

        Parameters
        ----------
        savefile : str
            Path of the file to save the basis to.
        overwrite : bool
            If ``True``, overwrite the file if it already exists.
            If ``False`` (default), raise a ``FileExistsError`` if the file
            already exists.
        """
        with utils.hdf5_savehandle(savefile, overwrite) as hf:
            if self.name:
                meta = hf.create_dataset("meta", shape=(0,))
                meta.attrs["name"] = self.name
            hf.create_dataset("entries", data=self.entries)
            if (w := self.weights) is not None:
                if isinstance(w, sparse.dia_array):
                    w = w.data[0]
                hf.create_dataset("weights", data=w)

    @classmethod
    def load(cls, loadfile: str):
        """Load a basis from an HDF5 file.

        Parameters
        ----------
        loadfile : str
            Path to the file where the basis was stored via :meth:`save`.

        Returns
        -------
        LinearBasis
        """
        with utils.hdf5_loadhandle(loadfile) as hf:
            name = None
            if "meta" in hf:
                name = hf["meta"].attrs["name"]
            entries = hf["entries"][:]
            weights = hf["weights"][:] if "weights" in hf else None
            return cls(entries, weights, name=name)



# lstsq/_total.py
"""Operator Inference truncated singular value decomposition solver."""

__all__ = [
    "TruncatedSVDSolver",
]

import types
import warnings
import numpy as np
import scipy.linalg as la

from .. import errors, utils
from ._base import SolverTemplate, _require_trained


class TruncatedSVDSolver(SolverTemplate):
    r"""Solve the Frobenius-norm ordinary least-squares problem with a
    low-rank best approximation of the data matrix.

    That is, solve

    .. math::
       \argmin_{\Ohat}\|\tilde{\D}\Ohat\trp - \Z\trp\|_{F}^{2}

    where :math:`\tilde{\D}` is the best rank-:math:`d'` approximation of
    :math:`\D` for a given :math:`d' \le \operatorname{rank}(\D)`, i.e.,

    .. math ::
       \tilde{\D}
       = \argmin_{\D' \in \RR^{k \times d}}\|\D' - \D\|_{F}
       \quad\textrm{such that}\quad
       \operatorname{rank}(\D') = d'.

    If :math:`\D = \bfPhi\bfSigma\bfPsi\trp` is the singular value
    decomposition of :math:`\D`, then defining

    .. math::
       \bfPhi' = \bfPhi_{:d', :}
       \quad
       \bfSigma' = \bfSigma_{:d', :d'}
       \quad
       \bfPsi' = \bfPsi_{:d', :},

    the solution to this problem is given by
    :math:`\Ohat\trp = \bfPsi'(\bfSigma')^{-1}(\bfPhi')\trp\Z\trp` (i.e.,
    :math:`\Ohat = \Z\bfPhi'(\bfSigma')^{-1}(\bfPsi')\trp`).

    Parameters
    ----------
    num_svdmodes : int or None
        Number of singular value decomposition modes to retain.
        If ``None``, use all available modes (no truncation).
        If a negative integer, use ``k - abs(num_svdmodes)`` modes where
        ``k`` is the number of available modes.
    lapack_driver : str
        LAPACK routine for computing the singular value decomposition.
        See :func:`scipy.linalg.svd()`.
    """

    def __init__(self, num_svdmodes: int, lapack_driver: str = "gesdd"):
        """Set the number of SVD columns to keep."""
        SolverTemplate.__init__(self)
        self.num_svdmodes = num_svdmodes
        self.__options = types.MappingProxyType(
            dict(full_matrices=False, lapack_driver=lapack_driver)
        )

    # Properties --------------------------------------------------------------
    @property
    def num_svdmodes(self) -> int:
        """Number of singular value decomposition modes to retain."""
        return self.__nmodes

    @num_svdmodes.setter
    def num_svdmodes(self, nmodes: int):
        """Set the number of columns."""
        if nmodes is None:
            self.__nmodes = self.max_modes
            return

        if not isinstance(nmodes, int):
            raise TypeError("num_svdmodes must be an integer")

        if self.data_matrix is not None:
            d = self.max_modes
            if nmodes <= 0:
                nmodes = d + nmodes
            if nmodes > d:
                raise ValueError(f"only {d} SVD modes available")
        self.__nmodes = nmodes

    @property
    def options(self):
        """Keyword arguments for :func:`scipy.linalg.svd()`.
        These cannot be changed after instantiation.
        """
        return self.__options

    @property
    def max_modes(self):
        """Maximum number of singular value decomposition modes available."""
        return None if self.data_matrix is None else self._ZPhi.shape[1]

    def __str__(self):
        """String representation: dimensions + solver options."""
        start = SolverTemplate.__str__(self)
        kwargs = self._print_kwargs(self.options)
        out = [f"SVD solver: scipy.linalg.svd({kwargs})"]
        if self.data_matrix is not None:
            out.append(
                f"using {self.num_svdmodes} of {self.max_modes} SVD modes"
            )
            lines = start.split("\n")
            lines.insert(
                3,
                f"    Truncation condition number: {self.tcond():.4e}",
            )
            start = "\n".join(lines)
        return f"{start}\n  " + "\n  ".join(out)

    # Main methods ------------------------------------------------------------
    def fit(self, data_matrix: np.ndarray, lhs_matrix: np.ndarray):
        """Save data matrices and prepare to solve the regression problem."""
        SolverTemplate.fit(self, data_matrix, lhs_matrix)

        Phi, svals, PsiT = la.svd(self.data_matrix, **self.options)
        self._svals = svals
        self._ZPhi = self.lhs_matrix @ Phi
        self._SinvPsiT = PsiT / svals.reshape((-1, 1))

        # Reset num_svdmodes.
        d = self._ZPhi.shape[1]
        if (n := self.num_svdmodes) is None:
            n = d
        if n <= 0:
            n = d + n
        if n > d:
            warnings.warn(
                f"only {d} SVD modes available, "
                f"setting num_svdmodes={d} (was {n})",
                errors.OpInfWarning,
            )
            n = d
        self.__nmodes = n

        return self

    @_require_trained
    def solve(self) -> np.ndarray:
        r"""Solve the Operator Inference regression.

        Returns
        -------
        Ohat : (r, d) ndarray
            Operator matrix :math:`\Ohat` (not its transpose!).
        """
        n = self.num_svdmodes
        Ohat = self._ZPhi[:, :n] @ self._SinvPsiT[:n, :]
        return Ohat

    # Post-processing ---------------------------------------------------------
    @_require_trained
    def tcond(self) -> float:
        r"""Compute the :math:`2`-norm condition number of :math:`\tilde{\D}`,
        the low-rank approximation to :math:`\D` from the truncated SVD.
        """
        return self._svals[0] / self._svals[self.num_svdmodes - 1]

    # Persistence -------------------------------------------------------------
    def save(self, savefile: str, overwrite: bool = False):
        """Serialize the solver, saving it in HDF5 format.
        The model can be recovered with the :meth:`load()` class method.

        Parameters
        ----------
        savefile : str
            File to save to, with extension ``.h5`` (HDF5).
        overwrite : bool
            If ``True`` and the specified ``savefile`` already exists,
            overwrite the file. If ``False`` (default) and the specified
            ``savefile`` already exists, raise an error.
        """
        with utils.hdf5_savehandle(savefile, overwrite) as hf:
            self._save_dict(hf, "options")
            hf.create_dataset("num_svdmodes", data=[self.num_svdmodes])
            if self.data_matrix is not None:
                hf.create_dataset("data_matrix", data=self.data_matrix)
                hf.create_dataset("lhs_matrix", data=self.lhs_matrix)
                for attr in ("_svals", "_ZPhi", "_SinvPsiT"):
                    hf.create_dataset(attr, data=getattr(self, attr))

    @classmethod
    def load(cls, loadfile: str):
        """Load a serialized solver from an HDF5 file, created previously from
        the :meth:`save()` method.

        Parameters
        ----------
        loadfile : str
            Path to the file where the model was stored via :meth:`save()`.

        Returns
        -------
        solver : TruncatedSVDSolver
            Loaded solver.
        """
        with utils.hdf5_loadhandle(loadfile) as hf:
            options = cls._load_dict(hf, "options")
            num_svdmodes = int(hf["num_svdmodes"][0])
            solver = cls(num_svdmodes, lapack_driver=options["lapack_driver"])
            if "data_matrix" in hf:
                D = hf["data_matrix"][:]
                Z = hf["lhs_matrix"][:]
                SolverTemplate.fit(solver, D, Z)
                for attr in ("_svals", "_ZPhi", "_SinvPsiT"):
                    setattr(solver, attr, hf[attr][:])
        return solver

    def copy(self):
        """Make a copy of the solver."""
        solver = self.__class__(
            num_svdmodes=self.num_svdmodes,
            lapack_driver=self.options["lapack_driver"],
        )
        if self.data_matrix is not None:
            SolverTemplate.fit(solver, self.data_matrix, self.lhs_matrix)
            for attr in ("_svals", "_ZPhi", "_SinvPsiT"):
                setattr(solver, attr, getattr(self, attr))
        return solver



# models/mono/_base.py
"""Abstract base class for monolithic dynamical systems models."""

__all__ = []

import abc
import warnings
import numpy as np

from ... import errors, lstsq, operators as _operators
from ...operators import _utils as oputils


class _Model(abc.ABC):
    """Base class for all monolithic models.

    Child classes:

    * :class:`opinf.models.mono._nonparametric._NonparametricModel`
    * :class:`opinf.models.mono._parametric._ParametricModel`
    """

    def __init__(self, operators, solver=None):
        """Define the model structure.

        Parameters
        ----------
        operators : list of :mod:`opinf.operators` objects
            Operators comprising the terms of the model.
        solver : :mod:`opinf.lstsq` object
            Solver for the Operator Inference regression.
        """
        self.__r = None
        self.__m = None
        self.__operators = None
        self.__solver = None

        self.operators = operators
        self.solver = solver

    # Properties: operators ---------------------------------------------------
    _operator_abbreviations = dict()  # Abbreviations for model operators.

    @staticmethod
    @abc.abstractmethod
    def _isvalidoperator(op):  # pragma: no cover
        """Return True if and only if ``op`` is a valid operator object
        for this class of model.
        """
        raise NotImplementedError

    @staticmethod
    @abc.abstractmethod
    def _check_operator_types_unique(ops):  # pragma: no cover
        """Raise a ValueError if any two operators represent the same kind
        of operation (e.g., two constant operators).
        """
        raise NotImplementedError

    @abc.abstractmethod
    def _get_operator_of_type(self, OpClass):  # pragma: no cover
        """Return the first operator in the model corresponding to the
        operator class ``OpClass``.
        """
        raise NotImplementedError

    @property
    def operators(self):
        """Operators comprising the terms of the model."""
        return self.__operators

    @operators.setter
    def operators(self, ops):
        """Set the operators."""
        if self._isvalidoperator(ops):
            ops = [ops]
        if len(ops) == 0 or ops is None:
            raise ValueError("at least one operator required")

        toinfer = []  # Operators to infer (no entries yet).
        known = []  # Operators whose entries are set.
        self._has_inputs = False  # Whether any operators use inputs.
        ops = list(ops)
        for i in range(len(ops)):
            op = ops[i]
            if isinstance(op, str):
                if op in (self._operator_abbreviations):
                    op = ops[i] = self._operator_abbreviations[op]()
                else:
                    raise TypeError(
                        f"operator abbreviation '{op}' not recognized"
                    )
            if not self._isvalidoperator(op):
                raise TypeError(
                    f"invalid operator of type '{op.__class__.__name__}'"
                )
            if oputils.is_uncalibrated(op):
                toinfer.append(i)
            else:
                known.append(i)
            if oputils.has_inputs(op):
                self._has_inputs = True
        self._check_operator_types_unique([ops[i] for i in toinfer])

        # Store attributes.
        self.__r = self._check_state_dimension_consistency(ops)
        self.__m = self._check_input_dimension_consistency(ops)
        self.__operators = ops
        self._indices_of_operators_to_infer = toinfer
        self._indices_of_known_operators = known
        self._fully_intrusive = len(toinfer) == 0

    def _clear(self):
        """Reset the entries of the non-intrusive operators and the
        state and input dimensions.
        """
        for i in self._indices_of_operators_to_infer:
            self.operators[i]._clear()
        self.__r = self._check_state_dimension_consistency(self.operators)
        self.__m = self._check_input_dimension_consistency(self.operators)

    def __iter__(self):
        """Iterate through the model operators."""
        for op in self.operators:
            yield op

    @property
    def c_(self):
        """:class:`opinf.operators.ConstantOperator` (or ``None``)."""
        return self._get_operator_of_type(_operators.ConstantOperator)

    @property
    def A_(self):
        """:class:`opinf.operators.LinearOperator` (or ``None``)."""
        return self._get_operator_of_type(_operators.LinearOperator)

    @property
    def H_(self):
        """:class:`opinf.operators.QuadraticOperator` (or ``None``)."""
        return self._get_operator_of_type(_operators.QuadraticOperator)

    @property
    def G_(self):
        """:class:`opinf.operators.CubicOperator` (or ``None``)."""
        return self._get_operator_of_type(_operators.CubicOperator)

    @property
    def B_(self):
        """:class:`opinf.operators.InputOperator` (or ``None``)."""
        return self._get_operator_of_type(_operators.InputOperator)

    @property
    def N_(self):
        """:class:`opinf.operators.StateInputOperator` (or ``None``)."""
        return self._get_operator_of_type(_operators.StateInputOperator)

    # Properties: dimensions --------------------------------------------------
    @staticmethod
    def _check_state_dimension_consistency(ops):
        """Ensure all operators have the same state dimension, except for
        inferrable operators whose entries have not been set.
        """
        rs = {
            op.state_dimension for op in ops if not oputils.is_uncalibrated(op)
        }
        if len(rs) > 1:
            raise errors.DimensionalityError(
                "operators not aligned "
                "(state dimension must be the same for all operators)"
            )
        return rs.pop() if len(rs) == 1 else None

    @property
    def state_dimension(self):
        """Dimension :math:`r` of the state."""
        return self.__r

    @state_dimension.setter
    def state_dimension(self, r):
        """Set the state dimension.
        Not allowed if any existing operators have ``state_dimension != r``.
        """
        if self.__operators is not None:
            for op in self.operators:
                if op.entries is not None and op.state_dimension != r:
                    raise AttributeError(
                        "can't set attribute "
                        f"(existing operators have r = {self.__r})"
                    )
        self.__r = int(r)

    @staticmethod
    def _check_input_dimension_consistency(ops):
        """Ensure all *input* operators with initialized entries have the same
        ``input dimension``.
        """
        inputops = [op for op in ops if oputils.has_inputs(op)]
        if len(inputops) == 0:
            return 0
        ms = {
            op.input_dimension
            for op in inputops
            if not oputils.is_uncalibrated(op)
        }
        if len(ms) > 1:
            raise errors.DimensionalityError(
                "input operators not aligned "
                "(input dimension must be the same for all input operators)"
            )
        return ms.pop() if len(ms) == 1 else None

    @property
    def input_dimension(self):
        """Dimension :math:`m` of the input (zero if there are no inputs)."""
        return self.__m

    @input_dimension.setter
    def input_dimension(self, m):
        """Set the input dimension.
        Only allowed if an input-using operator is present in the model
        and the ``input_dimension`` of every existing input operator agrees.
        """
        if not self._has_inputs and m != 0:
            raise AttributeError("can't set attribute (no input operators)")
        if self.__operators is not None:
            for op in self.operators:
                if (
                    oputils.has_inputs(op)
                    and not oputils.is_uncalibrated(op)
                    and op.input_dimension != m
                ):
                    raise AttributeError(
                        "can't set attribute "
                        f"(existing input operators have m = {self.__m})"
                    )
        self.__m = int(m)

    # Properties: solver ------------------------------------------------------
    @property
    def solver(self):
        """Solver for the least-squares regression, see :mod:`opinf.lstsq`."""
        return self.__solver

    @solver.setter
    def solver(self, solver):
        """Set the solver, including default options."""
        if self._fully_intrusive:
            if solver is not None:
                warnings.warn(
                    "all operators initialized explicity, setting solver=None",
                    errors.OpInfWarning,
                )
            self.__solver = None
            return

        # Defaults and shortcuts.
        if solver is None:
            # No regularization.
            solver = lstsq.PlainSolver()
        elif np.isscalar(solver):
            if solver == 0:
                # Also no regularization.
                solver = lstsq.PlainSolver()
            elif solver > 0:
                # Scalar Tikhonov (L2) regularization.
                solver = lstsq.L2Solver(solver)
            else:
                raise ValueError("if a scalar, solver must be nonnegative")

        # Light validation: must be instance w/ fit(), solve().
        if isinstance(solver, type):
            raise TypeError("solver must be an instance, not a class")
        for mtd in "fit", "solve":
            if not hasattr(solver, mtd) or not callable(getattr(solver, mtd)):
                warnings.warn(
                    f"solver should have a '{mtd}()' method",
                    errors.OpInfWarning,
                )

        self.__solver = solver

    # Dimensionality reduction ------------------------------------------------
    def galerkin(self, Vr, Wr=None):
        r"""Construct a reduced-order model by taking the (Petrov-)Galerkin
        projection of each model operator.

        Consider a model :math:`\z = \f(\q, \u)` where

        * :math:`\q\in\RR^n` is the model state,
        * :math:`\u\in\RR^m` is the input, and
        * :math:`\z\in\RR^n` is the model left-hand side.

        Given a *trial basis* :math:`\Vr\in\RR^{n\times r}` and a *test basis*
        :math:`\Wr\in\RR^{n\times r}`, the corresponding
        *intrusive reduced-order model* is the model
        :math:`\zhat = \fhat(\qhat, \u)` where

        .. math::
           \zhat = \Wr\trp\z,
           \qquad
           \fhat(\qhat,\u) = (\Wr\trp\Vr)^{-1}\Wr\trp\f(\Vr\qhat,\u).

        Here,

        * :math:`\qhat\in\RR^r` is the reduced-order state,
        * :math:`\u\in\RR^m` is the input (as before), and
        * :math:`\zhat\in\RR^r` is the reduced-order left-hand side.

        This approach uses the low-dimensional state approximation
        :math:`\q = \Vr\qhat`.
        If :math:`\Wr = \Vr`, the result is called a *Galerkin projection*.
        If :math:`\Wr \neq \Vr`, it is called a *Petrov-Galerkin projection*.

        Parameters
        ----------
        Vr : (n, r) ndarray
            Basis for the trial space.
        Wr : (n, r) ndarray or None
            Basis for the test space. If ``None``, defaults to ``Vr``.

        Returns
        -------
        reduced_model : Model
            Reduced-order model obtained from (Petrov-)Galerkin projection.
        """
        # TODO: preserve the _to_infer indices?
        return self.__class__(
            [
                (
                    old_op.copy()
                    if oputils.is_uncalibrated(old_op)
                    else old_op.galerkin(Vr, Wr)
                )
                for old_op in self.operators
            ]
        )

    # Validation methods ------------------------------------------------------
    def _check_inputargs(self, u, argname):
        """Check that the model structure agrees with input arguments."""
        if self._has_inputs and u is None:
            raise ValueError(f"argument '{argname}' required")

        if not self._has_inputs and u is not None:
            warnings.warn(
                f"argument '{argname}' should be None, "
                "argument will be ignored",
                errors.OpInfWarning,
            )

    def _check_is_trained(self):
        """Ensure that the model is trained and ready for prediction."""
        if self.state_dimension is None:
            raise AttributeError("no state_dimension (call fit())")
        if self._has_inputs and (self.input_dimension is None):
            raise AttributeError("no input_dimension (call fit())")
        if any(oputils.is_uncalibrated(op) for op in self.operators):
            raise AttributeError("model not trained (call fit())")

    def __eq__(self, other):
        """Two models are equal if they have equivalent operators."""
        if not isinstance(other, self.__class__):
            return False
        if len(self.operators) != len(other.operators):
            return False
        if self.state_dimension != other.state_dimension:
            return False
        if self.input_dimension != other.input_dimension:
            return False
        marked = set()
        for selfop in self.operators:
            found = False
            for i, otherop in enumerate(other.operators):
                if selfop == otherop and i not in marked:
                    found = True
                    marked.add(i)
                    break
            if not found:
                return False
        return True

    # Model persistence -------------------------------------------------------
    def copy(self):
        """Make a copy of the model."""
        sol = self.solver.copy() if self.solver is not None else None
        return self.__class__([op.copy() for op in self.operators], solver=sol)



# utils/_requires.py
"""Wrappers for methods that require an attribute to be initialized."""

__all__ = [
    "requires",
    "requires2",
]

import typing
import functools


def requires2(attr: str, message: str) -> typing.Callable:
    """Wrapper for methods that require an attribute to be initialized.

    Parameters
    ----------
    attr : str
        Name of the required attribute.
    message : str
        Message in the error.
    """

    def _wrapper(func):
        @functools.wraps(func)
        def _decorator(self, *args, **kwargs):
            if not hasattr(self, attr) or getattr(self, attr) is None:
                raise AttributeError(message)
            return func(self, *args, **kwargs)

        return _decorator

    return _wrapper


def requires(attr: str) -> typing.Callable:
    """Wrapper for methods that require an attribute to be initialized.

    Parameters
    ----------
    attr : str
        Name of the required attribute.
    """

    return requires2(attr, f"required attribute '{attr}' not set")



# __init__.py
"""Operator Inference for data-driven model reduction of dynamical systems.

Author: Willcox Research Group
Maintainer: Shane A. McQuarrie
GitHub:
    https://github.com/Willcox-Research-Group/rom-operator-inference-Python3
"""

__version__ = "0.5.13"

from . import (
    basis,
    errors,
    ddt,
    lift,
    lstsq,
    models,
    operators,
    pre,
    post,
    roms,
    utils,
)

from .roms import ROM, ParametricROM

__all__ = [
    "basis",
    "errors",
    "ddt",
    "lift",
    "lstsq",
    "models",
    "operators",
    "pre",
    "post",
    "roms",
    "utils",
    "ROM",
    "ParametricROM",
]



# basis/_multi.py
"""Basis for states with multiple variables."""

__all__ = [
    "BasisMulti",
]

import warnings
import numpy as np

from .. import errors, utils
from ._base import BasisTemplate


class BasisMulti:
    r"""Join bases together for states with multiple variables.

    This class is for states that can be written (after discretization) as

    .. math::
       \q = \left[\begin{array}{c}
       \q_0 \\ \q_1 \\ \vdots \\ \q_{n_q - 1}
       \end{array}\right]
       \in \RR^n,

    where each :math:`\q_i \in \RR^{n_i}` represents a single discretized
    state variable, to be compressed (individually) to :math:`r_i` degrees of
    freedom. The compressed state is therefore

    .. math::
       \qhat = \left[\begin{array}{c}
       \qhat_0 \\ \qhat_1 \\ \vdots \\ \qhat_{n_q - 1}
       \end{array}\right]
       \in \RR^r,

    where each :math:`\qhat_i \in \RR^{r_i}` is the compressed version of the
    state variable :math:`\q_i \in \RR^{n_i}`.
    The full state dimension is :math:`n = \sum_{i=0}^{n_q - 1} n_i`, i.e.,
    ``full_state_dimension = sum(full_variable_sizes)``; the reduced state
    dimension is :math:`r = \sum_{i=0}^{n_q - 1}r_i`, i.e.,
    ``reduced_state_dimension = sum(reduced_variable_sizes)``.

    Parameters
    ----------
    bases : list
        Initialized (but not necessarily trained) basis objects,
        one for each state variable.
    full_variable_sizes : list or None
        Dimensions for each state variable, :math:`n_0,\ldots,n_{n_q-1}`.
        If ``None`` (default), set :math:`n_i` to
        ``bases[i].full_state_dimension``; if any of these are ``None``,
        assume all state variables have the same dimension, i.e.,
        :math:`n_0 = n_1 = \cdots = n_x \in \NN` with :math:`n_x` to be
        determined in :meth:`fit`. In this case, :math:`n = n_q n_x`.
    """

    def __init__(self, bases, full_variable_sizes=None):
        """Initialize the bases."""
        self.bases = bases

        if full_variable_sizes is not None:
            if len(full_variable_sizes) != len(bases):
                raise ValueError("len(full_variable_sizes) != len(bases)")
            for basis, ni in zip(bases, full_variable_sizes):
                BasisTemplate.full_state_dimension.fset(basis, ni)

    # Properties --------------------------------------------------------------
    @property
    def bases(self) -> tuple:
        """Bases for each state variable."""
        return self.__bases

    @bases.setter
    def bases(self, bs):
        """Set the bases."""
        if (num_variables := len(bs)) == 0:
            raise ValueError("at least one basis required")
        elif num_variables == 1:
            warnings.warn("only one variable detected", errors.OpInfWarning)

        # Check inheritance and set default variable names.
        for i, basis in enumerate(bs):
            if not isinstance(basis, BasisTemplate):
                warnings.warn(
                    f"bases[{i}] does not inherit from "
                    "BasisTemplate, unexpected behavior may occur",
                    errors.OpInfWarning,
                )
            if basis.name is None:
                basis.name = f"variable {i}"

        # Store bases and set slice dimensions..
        self.__nq = num_variables
        self.__bases = tuple(bs)

    @property
    def num_variables(self) -> int:
        r"""Number of state variables :math:`n_q \in \NN`."""
        return self.__nq

    @property
    def variable_names(self) -> tuple:
        """Names for each state variable."""
        return tuple(basis.name for basis in self.bases)

    @property
    def full_variable_sizes(self) -> tuple:
        r"""Dimensions of each state variable,
        :math:`n_0, \ldots, n_{n_q - 1} \in \NN`.
        """
        return tuple(basis.full_state_dimension for basis in self.bases)

    @property
    def full_state_dimension(self) -> int:
        r"""Total dimension :math:`n = \sum_{i=0}^{n_q-1} n_i \in \NN` of the
        joint full state.
        """
        if None in (ns := self.full_variable_sizes):
            return None
        return sum(ns)

    @property
    def reduced_variable_sizes(self) -> tuple:
        r"""Dimensions of each compressed state variable,
        :math:`r_0, \ldots, r_{n_q - 1} \in \NN`.
        """
        return tuple(basis.reduced_state_dimension for basis in self.bases)

    @property
    def reduced_state_dimension(self):
        r"""Total dimension :math:`r = \sum_{i=0}^{n_q - 1}r_i \in \NN`
        of the joint reduced state.
        """
        if None in (rs := self.reduced_variable_sizes):
            return None
        return sum(rs)

    @property
    def shape(self) -> tuple[int, int]:
        """Dimensions :math:`(n, r)` of the basis."""
        return (self.full_state_dimension, self.reduced_state_dimension)

    # Magic methods -----------------------------------------------------------
    def __len__(self) -> int:
        """Length = number of state variables."""
        return self.num_variables

    def __getitem__(self, key):
        """Get the basis for variable i."""
        if key in self.variable_names:
            key = self.variable_names.index(key)
        return self.bases[key]

    def __eq__(self, other) -> bool:
        """Test two BasisMulti objects for equality."""
        if not (
            isinstance(other, self.__class__)
            or isinstance(self, other.__class__)
        ):
            return False
        if self.num_variables != other.num_variables:
            return False
        return all(b1 == b2 for b1, b2 in zip(self.bases, other.bases))

    def __str__(self) -> str:
        """String representation: str() of each basis."""
        out = [f"{self.num_variables}-variable {self.__class__.__name__}"]
        out += [str(basis) for basis in self.bases]
        return "\n| ".join("\n\n".join(out).split("\n"))

    def __repr__(self) -> str:
        """Unique ID + string representation."""
        return utils.str2repr(self)

    # Access convenience methods ----------------------------------------------
    def _check_shape_full(self, states):
        """Verify the shape of the full states."""
        if (n := self.full_state_dimension) is None:
            raise AttributeError("full_state_dimension not set, call fit()")
        if (n2 := states.shape[0]) != n:
            raise errors.DimensionalityError(
                f"states.shape[0] = {n2:d} != {n:d} = full_state_dimension"
            )

    def _check_shape_reduced(self, states_compressed):
        """Verify the shape of the reduced states."""
        if (r := self.reduced_state_dimension) is None:
            raise AttributeError("reduced_state_dimension not set, call fit()")
        if (r2 := states_compressed.shape[0]) != r:
            raise errors.DimensionalityError(
                f"states_compressed.shape[0] = {r2:d} "
                f"!= {r:d} = reduced_state_dimension"
            )

    def _check_is_reduced(self, states) -> bool:
        """Verify the shape of the full or reduced vector / matrix Q.

        Parameters
        ----------
        states : (n, ...) or (r, ...) ndarray
            Joint full or reduced state vector / snapshot matrix.

        Returns
        -------
        is_reduced : bool
            ``False`` if ``states`` has `n` rows;
            ``True`` if ``states`` has `r` rows.

        Raises
        ------
        errors.DimensionalityError
            If states is neither `n`- nor `r`-dimensional.
        """
        if (n := self.full_state_dimension) is not None:
            if states.shape[0] == n:
                return False
        if (r := self.reduced_state_dimension) is not None:
            if states.shape[0] == r:
                return True

        if n is r is None:
            raise AttributeError("dimension attributes not set")
        raise errors.DimensionalityError(
            f"states.shape[0] must be "
            f"full_state_dimension (n = {n:d}) or "
            f"reduced_state_dimension (r = {r:d})"
        )

    def _slices(self, varindex: int = None, reduced: bool = False):
        """Get slices for one or all full or reduced state variable(s).

        Parameters
        ----------
        varindex : int
            Index of the variable to get a slice for.
            If ``None`` (default), get slices for all state variables.
        reduced : bool
            If ``True``, slice the reduced state variables.
            If ``False`` (default), slice the full state variables.

        Returns
        -------
        slice or tuple
            Slice for the full state variable at index ``varindex``
            or a tuple of slices for all state variables.
        """
        ds = (
            self.reduced_variable_sizes
            if reduced
            else self.full_variable_sizes
        )
        dimsum = np.cumsum((0,) + ds)

        if varindex is not None:
            return slice(dimsum[varindex], dimsum[varindex + 1])

        return [
            slice(dimsum[i], dimsum[i + 1]) for i in range(self.num_variables)
        ]

    def get_var(self, var, states):
        """Extract a single variable from the joint full or reduced state.

        Parameters
        ----------
        var : int or str
            Index or name of the variable to extract.
        states : (n, ...) or (r, ...) ndarray
            Joint full or reduced state vector / snapshot matrix.

        Returns
        -------
        state_variable : (n_i, ...) or (r_i, ...) ndarray
            One full or reduced state variable, extracted from ``states``.
        """
        if var in self.variable_names:
            var = self.variable_names.index(var)
        if self._check_is_reduced(states):
            return states[self._slices(var, reduced=True)]
        return states[self._slices(var, reduced=False)]

    def split(self, states):
        """Split a full or reduced state vector into the individual variables.

        Parameters
        ----------
        states : (r, ...) ndarray
            Joint full or reduced state vector / snapshot matrix to split.

        Returns
        -------
        arrs : list ``num_variables`` ndarrays
            Individual full or reduced state variables.
        """
        if self._check_is_reduced(states):
            return [states[s] for s in self._slices(reduced=True)]
        return [states[s] for s in self._slices(reduced=False)]

    # Main routines -----------------------------------------------------------
    def fit(self, states):
        """Construct the joint basis by calling ``fit()`` on the basis for
        each variable.

        Parameters
        ----------
        states : (n, k) ndarray
            Matrix of `k` `n`-dimensional snapshots.
            The first ``full_variable_sizes[0]`` entries correspond to the
            first state variable, the next ``full_variable_sizes[1]`` entries
            correspond to the second state variable, and so on.
            If ``full_variable_sizes`` are not yet prescribed, assume that each
            state variable has the same dimension.

        Returns
        -------
        self
        """
        if self.full_state_dimension is None:
            # Assume all state dimensions are equal.
            nx, remainder = divmod(states.shape[0], self.num_variables)
            if remainder != 0:
                raise errors.DimensionalityError(
                    "len(states) must be evenly divisible by "
                    f"the number of variables n_q = {self.num_variables}"
                )
            for basis in self.bases:
                BasisTemplate.full_state_dimension.fset(basis, nx)

        self._check_shape_full(states)
        for basis, Q in zip(self.bases, self.split(states)):
            basis.fit(Q)

        return self

    def compress(self, states):
        """Map high-dimensional states to low-dimensional latent coordinates.

        Parameters
        ----------
        states : (n, ...) ndarray
            Matrix of `n`-dimensional state vectors, or a single state vector.
            The first ``full_variable_sizes[0]`` entries correspond to the
            first state variable, the next ``full_variable_sizes[1]`` entries
            correspond to the second state variable, and so on.

        Returns
        -------
        states_compressed : (r, ...) ndarray
            Matrix of `r`-dimensional latent coordinate vectors, or a single
            coordinate vector. The first ``reduced_variable_sizes[0]``
            entries correspond to the first reduced state variable, the next
            ``reduced_variable_sizes[1]`` entries correspond to the second
            reduced state variable, and so on.
        """
        self._check_shape_full(states)

        return np.concatenate(
            [
                basis.compress(Q)
                for basis, Q in zip(self.bases, self.split(states))
            ],
            axis=0,
        )

    def decompress(self, states_compressed, locs=None):
        r"""Map low-dimensional latent coordinates to high-dimensional states.

        Parameters
        ----------
        states_compressed : (r, ...) ndarray
            Matrix of `r`-dimensional latent coordinate vectors, or a single
            coordinate vector. The first ``reduced_variable_sizes[0]``
            entries correspond to the first reduced state variable, the next
            ``reduced_variable_sizes[1]`` entries correspond to the second
            reduced state variable, and so on.
        locs : slice or (p,) ndarray of integers or None
            If given, decompress each state variable at *only* the
            `p` specified locations (indices) described by ``locs``.
            This option requires each full state variable to have the
            same dimension.

        Returns
        -------
        states_decompressed : (n, ...) or (num_variables*p, ...) ndarray
            Matrix of `n`-dimensional decompressed state vectors, or the
            :math:`n_q \cdot p` entries of such at the entries specified
            by ``locs``.
        """
        self._check_shape_reduced(states_compressed)
        if locs is not None and len(set(self.full_variable_sizes)) > 1:
            raise ValueError(
                "'locs != None' requires that "
                "all bases have the same full_state_dimension"
            )

        return np.concatenate(
            [
                basis.decompress(Q_, locs=locs)
                for basis, Q_ in zip(self.bases, self.split(states_compressed))
            ]
        )

    def project(self, state):
        """Project a high-dimensional state vector to the subset of the
        high-dimensional space that can be represented by the basis.

        This is done by

        1. expressing the state in low-dimensional latent coordinates, then
        2. reconstructing the high-dimensional state corresponding to those
           coordinates.

        In other words, ``project(Q)`` is equivalent to
        ``decompress(compress(Q))``.

        Parameters
        ----------
        states : (n, ...) ndarray
            Matrix of `n`-dimensional state vectors, or a single state vector.
            The first ``full_variable_sizes[0]`` entries correspond to the
            first state variable, the next ``full_variable_sizes[1]`` entries
            correspond to the second state variable, and so on.

        Returns
        -------
        state_projected : (n, ...) ndarray
            Matrix of `n`-dimensional projected state vectors, or a single
            projected state vector. The first ``full_variable_sizes[0]``
            entries correspond to the first state variable, the next
            ``full_variable_sizes[1]`` entries correspond to the second state
            variable, and so on.
        """
        return self.decompress(self.compress(state))

    def projection_error(self, state, relative=True) -> float:
        r"""Compute the error of the basis representation of a state or states.

        This function computes :math:`\frac{\|\Q - \mathcal{P}(\Q)\|}{\|\Q\|}`,
        where :math:`\Q` is the ``state`` and :math:`\mathcal{P}` is the
        projection defined by :meth:`project()`.
        If ``state`` is one-dimensional then :math:`||\cdot||` is the vector
        2-norm. If ``state`` is two-dimensional then :math:`||\cdot||` is the
        Frobenius norm.

        Parameters
        ----------
        state : (n,) or (n, k) ndarray
            Matrix of `n`-dimensional state vectors, or a single state vector.
            The first ``full_variable_sizes[0]`` entries correspond to the
            first state variable, the next ``full_variable_sizes[1]`` entries
            correspond to the second state variable, and so on.
        relative : bool
            If ``True`` (default), return the relative projection error
            ``norm(state - project(state)) / norm(state)``.
            If ``False``, return the absolute projection error
            ``norm(state - project(state))``.

        Returns
        -------
        float
            Relative error of the projection (``relative=True``) or
            absolute error of the projection (``relative=False``).
        """
        return BasisTemplate.projection_error(self, state, relative)

    # Model persistence -------------------------------------------------------
    def save(self, savefile, overwrite=False):
        """Save the transformer to an HDF5 file.

        Parameters
        ----------
        savefile : str
            Path of the file to save the transformer to.
        overwrite : bool
            If ``True``, overwrite the file if it already exists. If ``False``
            (default), raise a ``FileExistsError`` if the file already exists.
        """
        with utils.hdf5_savehandle(savefile, overwrite) as hf:
            hf.create_dataset("num_variables", data=[self.num_variables])
            for i, basis in enumerate(self.bases):
                basis.save(hf.create_group(f"variable{i}"))

    @classmethod
    def load(cls, loadfile, BasisClasses):
        """Load a previously saved transformer from an HDF5 file.

        Parameters
        ----------
        loadfile : str
            File where the transformer was stored via :meth:`save()`.
        BasisClasses : Iterable[type]
            Classes of the bases for each state variable.

        Returns
        -------
        TransformerMulti
        """
        with utils.hdf5_loadhandle(loadfile) as hf:
            num_variables = int(hf["num_variables"][0])

            if isinstance(BasisClasses, type):
                BasisClasses = [BasisClasses] * num_variables
            if (nclasses := len(BasisClasses)) != num_variables:
                raise ValueError(
                    f"file contains {num_variables:d} bases "
                    f"but {nclasses:d} classes provided"
                )

            bases = [
                BasisClasses[i].load(hf[f"variable{i}"])
                for i in range(num_variables)
            ]

            return cls(bases)

    # Verification ------------------------------------------------------------
    def verify(self):
        """Verify that :meth:`compress()` and :meth:`decompress()` are
        consistent in the sense that the range of :meth:`decompress()` is in
        the domain of :meth:`compress()` and that :meth:`project()` defines
        a projection operator, i.e., ``project(project(Q)) = project(Q)``.
        """
        for basis, name in zip(self.bases, self.variable_names):
            print(f"{name}:", end="\t")
            basis.verify()
        BasisTemplate.verify(self)

    def _verify_locs(self, states_compressed, states_projected):
        """Verify :meth:`decompress()` with ``locs != None``."""
        if len(sizes := set(self.full_variable_sizes)) != 1:
            return  # Cannot use locs unless all full variable sizes equal.
        nx = sizes.pop()
        locs = np.sort(np.random.choice(nx, size=(nx // 3), replace=False))

        states_projected_at_locs = np.concatenate(
            [Q[locs] for Q in self.split(states_projected)]
        )
        states_at_locs_projected = self.decompress(
            states_compressed,
            locs=locs,
        )

        if states_at_locs_projected.shape != states_projected_at_locs.shape:
            raise errors.VerificationError(  # pragma: no cover
                "decompress(states_compressed, locs).shape "
                "!= decompressed_states_at_locs.shape"
            )
        if not np.allclose(states_at_locs_projected, states_projected_at_locs):
            raise errors.VerificationError(  # pragma: no cover
                "decompress(states_compressed, locs) "
                "!= decompressed_states_at_locs"
            )



# errors.py
"""Custom exception classes."""


class DimensionalityError(ValueError):  # pragma: no cover
    """Dimension of data not aligned with previous model information."""

    pass


class LoadfileFormatError(Exception):  # pragma: no cover
    """File format inconsistent with a loading routine."""

    pass


class VerificationError(RuntimeError):  # pragma: no cover
    """Implementation of a template fails to meet requriements."""

    pass


class OpInfWarning(UserWarning):  # pragma: no cover
    """Generic warning for package usage."""

    pass



# operators/_affine.py
"""Classes for parametric OpInf operators where the parametric dependence is
expressed as an affine expansion.
"""

__all__ = [
    "AffineConstantOperator",
    "AffineLinearOperator",
    "AffineQuadraticOperator",
    "AffineCubicOperator",
    "AffineInputOperator",
    "AffineStateInputOperator",
]

import h5py
import warnings
import numpy as np
import scipy.sparse as sparse

from .. import errors, utils
from ._base import ParametricOpInfOperator, InputMixin
from ._nonparametric import (
    ConstantOperator,
    LinearOperator,
    QuadraticOperator,
    CubicOperator,
    InputOperator,
    StateInputOperator,
)


# Helper functions ============================================================
def _identity(x):
    """Identity function."""
    return x


def _is_iterable(obj):
    """Return True if obj is iterable, False, else."""
    try:
        iter(obj)
        return True
    except TypeError:
        return False


def _vectorizer(functions):
    """Translate a tuple of functions into a ndarray-valued function."""
    if any(not callable(func) for func in functions):
        raise TypeError("if 'coeffs' is iterable each entry must be callable")

    def _vectorized(parameter):
        return np.array([func(parameter) for func in functions])

    return _vectorized


# Base class ==================================================================
class _AffineOperator(ParametricOpInfOperator):
    r"""Base class for parametric operators where the parameter dependence
    can be written as an affine expansion with known scalar coefficients
    which are a function of the parameter vector.

    This type of operator can be written as

    .. math::
       \Ophat_{\ell}(\qhat,\u;\bfmu) = \left(\sum_{a=0}^{A_{\ell}-1}
       \theta_{\ell}^{(0)}\!(\bfmu)\Ohat_{\ell}^{(a)}
       \right)\d_{\ell}(\qhat, \u)

    where each :math:`\theta_{\ell}^{(a)}:\RR^{p}\to\RR` is a scalar-valued
    function of the parameter vector, each
    :math:`\Ohat_{\ell}^{(a)}\in\RR^{r\times d}` is a constant matrix, and
    :math:`\d:\RR^{r}\times\RR^{m}\to\RR^{d}.`

    Parent class: :class:`opinf.operators.ParametricOpInfOperator`

    Parameters
    ----------
    coeffs : callable, (iterable of callables), or int
        Coefficient functions for the terms of the affine expansion.

        * If callable, it should receive a parameter vector
          :math:`\bfmu` and return the vector of affine coefficients,
          :math:`[~\theta_{\ell}^{(0)}(\bfmu)
          ~~\cdots~~\theta_{\ell}^{(A_{\ell}-1)}(\bfmu)~]\trp`.
          In this case, ``nterms`` is a required argument.
        * If an iterable, each entry should be a callable representing a
          single affine coefficient function :math:`\theta_{\ell}^{(a)}`.
        * If an integer :math:`p`, set :math:`A_{\ell} = p` and define
          :math:`\theta_{\ell}^{(i)}\!(\bfmu) = \mu_i`. This is equivalent to
          using ``coeffs=lambda mu: mu``, except the parameter dimension is
          also captured and ``nterms`` is not required.
    nterms : int or None
        Number of terms :math:`A_{\ell}` in the affine expansion.
        Only required if ``coeffs`` is provided as a callable.
    entries : (list of ndarrays), ndarray, or None
        Operator matrices for each term of the affine expansion, i.e.,
        :math:`\Ohat_{\ell}^{(0)},\ldots,\Ohat_{\ell}^{(A_{\ell}-1)}.`
        If not provided in the constructor, use :meth:`set_entries()` later.
    fromblock : bool
        If ``True``, interpret ``entries`` as a horizontal concatenation
        of arrays; if ``False`` (default), interpret ``entries`` as a list
        of arrays.
    """

    # Initialization ----------------------------------------------------------
    def __init__(
        self,
        coeffs,
        nterms: int = None,
        entries=None,
        fromblock: bool = False,
    ):
        """Set coefficient functions and (if given) operator matrices."""
        ParametricOpInfOperator.__init__(self)
        if nterms is not None and (not isinstance(nterms, int) or nterms < 1):
            raise TypeError(
                "when provided, argument 'nterms' must be a positive integer"
            )
        self.__nterms = nterms

        # Parse the coefficient functions.
        if isinstance(coeffs, int) and coeffs > 0:
            if nterms is not None and nterms != coeffs:
                warnings.warn(
                    f"{coeffs} = coeffs != nterms = {nterms}, ignoring "
                    f"argument 'nterms' and setting nterms = {coeffs}",
                    errors.OpInfWarning,
                )
            self.__nterms = coeffs
            self.parameter_dimension = coeffs
            coeffs = _identity
        if not callable(coeffs):
            if not _is_iterable(coeffs):
                raise TypeError(
                    "argument 'coeffs' must be "
                    "callable, iterable, or a positive int"
                )
            A_ell = len(coeffs)
            if nterms is not None and nterms != A_ell:
                warnings.warn(
                    f"{A_ell} = len(coeffs) != nterms = {nterms}, ignoring "
                    f"argument 'nterms' and setting nterms = {A_ell}",
                    errors.OpInfWarning,
                )
            self.__nterms = A_ell
            coeffs = _vectorizer(coeffs)
        if self.__nterms is None:
            raise ValueError(
                "argument 'nterms' required when argument 'coeffs' is callable"
            )
        self.__thetas = coeffs

        if entries is not None:
            self.set_entries(entries, fromblock=fromblock)

    # Properties --------------------------------------------------------------
    def coeffs(self, parameter):
        r"""Evaluate the coefficient functions for each term of the affine
        expansion for a given parameter vector.

        This method represents the vector-valued function
        :math:`\boldsymbol{\theta}_{\ell} : \RR^{p} \to \RR^{A_{\ell}}`
        given by :math:`\boldsymbol{\theta}_{\ell}(\bfmu) = [~
        \theta_{\ell}^{(0)}~~\cdots~~\theta_{\ell}^{(A_{\ell}-1)}~]\trp.`

        Parameters
        ----------
        parameter : (p,) ndarray
            Parameter vector to evaluate.

        Returns
        -------
        coefficients : (nterms,) ndarray
            Coefficients of the affine expansion at the given ``parameter``.
        """
        return self.__thetas(parameter)

    @property
    def entries(self) -> list:
        r"""Operator matrices for each term of the affine expansion, i.e.,
        :math:`\Ohat_{\ell}^{(0)},\ldots,\Ohat_{\ell}^{(A_{\ell}-1)}.`
        """
        return ParametricOpInfOperator.entries.fget(self)

    @property
    def nterms(self) -> int:
        r"""Number of terms :math:`A_{\ell}` in the affine expansion."""
        return self.__nterms

    def set_entries(self, entries, fromblock: bool = False) -> None:
        r"""Set the operator matrices for each term of the affine expansion.

        Parameters
        ----------
        entries : list of s (r, d) ndarrays, or (r, sd) ndarray
            Operator matrices, either as a list of arrays
            (``fromblock=False``, default)
            or as a horizontal concatenatation of arrays (``fromblock=True``).
        fromblock : bool
            If ``True``, interpret ``entries`` as a horizontal concatenation
            of arrays; if ``False`` (default), interpret ``entries`` as a list
            of arrays.
        """
        # Extract / verify the entries.
        nterms = self.nterms
        if fromblock:
            if not isinstance(entries, np.ndarray) or (
                entries.ndim not in (1, 2)
            ):
                raise ValueError(
                    "entries must be a 1- or 2-dimensional ndarray "
                    "when fromblock=True"
                )
            entries = np.split(entries, nterms, axis=-1)
        if np.ndim(entries) > 1:
            self._check_shape_consistency(entries, "entries")
        if (n_arrays := len(entries)) != nterms:
            raise ValueError(
                f"{nterms} = number of affine expansion terms "
                f"!= len(entries) = {n_arrays}"
            )

        ParametricOpInfOperator.set_entries(
            self,
            [self._OperatorClass(A).entries for A in entries],
        )

    def __str__(self) -> str:
        lines = ParametricOpInfOperator.__str__(self).split("\n")
        lines.insert(-1, f"  expansion terms:     {self.nterms}")
        return "\n".join(lines)

    # Evaluation --------------------------------------------------------------
    @utils.requires("entries")
    def evaluate(self, parameter):
        r"""Evaluate the operator at the given parameter value.

        Parameters
        ----------
        parameter : (p,) ndarray or float
            Parameter value :math:`\bfmu` at which to evalute the operator.

        Returns
        -------
        op : nonparametric :mod:`opinf.operators` operator
            Nonparametric operator corresponding to the parameter value.
        """
        if self.parameter_dimension is None:
            self._set_parameter_dimension_from_values([parameter])
        self._check_parametervalue_dimension(parameter)
        theta_mus = self.coeffs(parameter)
        if self.nterms == 1 and np.isscalar(theta_mus):
            theta_mus = [theta_mus]
        entries = sum([tm * A for tm, A in zip(theta_mus, self.entries)])
        return self._OperatorClass(entries)

    # Dimensionality reduction ------------------------------------------------
    @utils.requires("entries")
    def galerkin(self, Vr, Wr=None):
        r"""Project this operator to a low-dimensional linear space.

        Consider an affine operator

        .. math::
            \Op_{\ell}(\q,\u;\bfmu)
            = \sum_{a=0}^{A_{\ell}-1}\theta_{\ell}^{(a)}\!(\bfmu)\,
            \Op_{\ell}^{(a)}\!(\q, \u)

        where

        * :math:`\q\in\RR^n` is the full-order state,
        * :math:`\u\in\RR^m` is the input,
        * :math:`\bfmu\in\RR^p` is the parameter vector, and
        * each :math:`\Op_{\ell}^{(a)}\!(\q,\u)` is a nonparametric operator.

        Given a *trial basis* :math:`\Vr\in\RR^{n\times r}` and a *test basis*
        :math:`\Wr\in\RR^{n\times r}`, the corresponding *intrusive projection*
        of :math:`\f` is the affine operator

        .. math::
           \fhat_{\ell}(\qhat,\u;\bfmu)
           = \sum_{a=0}^{A_{\ell}-1}\theta_{\ell}^{(a)}\!(\bfmu)\,
           (\Wr\trp\Vr)^{-1}\Wr\trp\Op_{\ell}^{(a)}\!(\V\qhat, \u)
           = \sum_{a=0}^{A_{\ell}-1}\theta_{\ell}^{(a)}\!(\bfmu)\,
           \Ophat_{\ell}^{(a)}\!(\qhat, \u),

        where :math:`\Ophat_{\ell}^{(a)}\!(\qhat, \u)
        = (\Wr\trp\Vr)^{-1}\Wr\trp\Op_{\ell}^{(a)}\!(\V\qhat, \u)`
        is the intrusive projection of :math:`\Op_{\ell}^{(a)}.`
        Here, :math:`\qhat\in\RR^r` is the reduced-order state, which enables
        the low-dimensional state approximation :math:`\q = \Vr\qhat.`
        If :math:`\Wr = \Vr`, the result is called a *Galerkin projection*.
        If :math:`\Wr \neq \Vr`, it is called a *Petrov-Galerkin projection*.

        Parameters
        ----------
        Vr : (n, r) ndarray
            Basis for the trial space.
        Wr : (n, r) ndarray or None
            Basis for the test space. If ``None``, defaults to ``Vr``.

        Returns
        -------
        op : operator
            New object of the same class as ``self``.
        """
        return self.__class__(
            coeffs=self.coeffs,
            nterms=self.nterms,
            entries=[
                self._OperatorClass(A).galerkin(Vr, Wr).entries
                for A in self.entries
            ],
            fromblock=False,
        )

    # Operator inference ------------------------------------------------------
    def operator_dimension(self, s: int, r: int, m: int) -> int:
        r"""Number of columns in the concatenated operator matrix.

        For affine operators, this is :math:`A_{\ell}\cdot d(r,m)`,
        where :math:`A_{\ell}` is the number of terms in the affine expansion
        and :math:`d(r,m)` is the dimension of the function
        :math:`\d(\qhat,\u)`.

        Parameters
        ----------
        s : int
            Number of training parameter values.
        r : int
            State dimension.
        m : int or None
            Input dimension.

        Returns
        -------
        d : int
            Number of columns in the concatenated operator matrix.
        """
        return self.nterms * self._OperatorClass.operator_dimension(r, m)

    def datablock(self, parameters, states, inputs=None) -> np.ndarray:
        r"""Return the data matrix block corresponding to the operator.

        For affine operators :math:`\Ophat_{\ell}(\qhat,\u;\bfmu)
        = \Ohat_{\ell}(\bfmu)\d_{\ell}(\qhat,\u)` with
        :math:`\Ohat_{\ell}(\bfmu)\in\RR^{r\times d}` and
        :math:`\d_{\ell}(\qhat,\u)\in\RR^{d}`, this is the block matrix

        .. math::
           \D_{\ell}\trp
           = \left[\begin{array}{ccc}
               \theta_{\ell}^{(0)}\!(\bfmu_{0})\,
               \d_{\ell}(\Qhat_{0},\U_{0})
               & \cdots &
               \theta_{\ell}^{(0)}\!(\bfmu_{s-1})\,
               \d_{\ell}(\Qhat_{s-1},\U_{s-1})
               \\ \vdots & & \vdots \\
               \theta_{\ell}^{(A_{\ell})}\!(\bfmu_{0})\,
               \d_{\ell}(\Qhat_{0},\U_{0})
               & \cdots &
               \theta_{\ell}^{(A_{\ell})}\!(\bfmu_{s-1})\,
               \d_{\ell}(\Qhat_{s-1},\U_{s-1})
           \end{array}\right]
           \in \RR^{A_{\ell}d \times \sum_{i=0}^{s-1}k_i}

        where :math:`\Qhat_{i} =
        [~\qhat_{i,0}~~\cdots~~\qhat_{i,k_i-1}] \in \RR^{r \times k_i}`
        and :math:`\U_{i} =
        [~\u_{i,0}~~\cdots~~\u_{i,k_i-1}] \in \RR^{m\times k_i}`
        are the state snapshots and inputs corresponding to training parameter
        value :math:`\bfmu_i\in\RR^{p}`, :math:`i = 0, \ldots, s-1`, where
        :math:`s` is the number of training parameter values. The notation
        :math:`\d_{\ell}(\Qhat_{i},\U_{i})` is shorthand for the matrix

        .. math::
           \d(\Qhat_{i},\U_{i})
           = \left[\begin{array}{ccc}
               \d_{\ell}(\qhat_{i,0},\u_{i,0})
               & \cdots &
               \d_{\ell}(\qhat_{i,k_i-1},\u_{i,k_i-1})
           \end{array}\right]
           \in \RR^{d \times k_i}.

        Parameters
        ----------
        parameters : (s, p) ndarray
            Traning parameter values :math:`\bfmu_{0},\ldots,\bfmu_{s-1}.`
        states : list of s (r, k) ndarrays
            State snapshots for each of the :math:`s` training parameter
            values, i.e., :math:`\Qhat_{0},\ldots,\Qhat_{s-1}.`
        inputs : list of s (m, k)-or-(k,) ndarrays or None
            Inputs corresponding to the state snapshots, i.e.,
            :math:`\U_{0},\ldots,\U_{s-1}.`
            If each input matrix is 1D, it is assumed that :math:`m = 1.`

        Returns
        -------
        block : (D, K) ndarray
            Data block for the affine operator. Here,
            :math:`D = A_{\ell}d(r,m)` and :math:`K = \sum_{i=0}^{s-1}k_i`
            is the total number of snapshots.
        """
        if not isinstance(self, InputMixin):
            inputs = [None] * len(parameters)
        blockcolumns = []
        for mu, Q, U in zip(parameters, states, inputs):
            Di = self._OperatorClass.datablock(Q, U)
            theta_mus = self.coeffs(mu)
            if self.nterms == 1 and np.isscalar(theta_mus):
                theta_mus = [theta_mus]
            blockcolumns.append(np.vstack([theta * Di for theta in theta_mus]))
        return np.hstack(blockcolumns)

    # Model persistence -------------------------------------------------------
    def copy(self):
        """Return a copy of the operator. Only the operator matrices are
        copied, not the coefficient functions.
        """
        As = None
        if self.entries is not None:
            As = [A.copy() for A in self.entries]
        op = self.__class__(
            coeffs=self.__thetas,
            nterms=self.nterms,
            entries=As,
            fromblock=False,
        )
        if self.parameter_dimension is not None:
            op.parameter_dimension = self.parameter_dimension
        return op

    def save(self, savefile: str, overwrite: bool = False) -> None:
        """Save the operator to an HDF5 file.

        Since the :attr:`coeffs` are callables, they cannot be
        serialized, and are therefore an argument to :meth:`load()`.

        Parameters
        ----------
        savefile : str
            Path of the file to save the basis in.
        overwrite : bool
            If ``True``, overwrite the file if it already exists. If ``False``
            (default), raise a ``FileExistsError`` if the file already exists.
        """
        with utils.hdf5_savehandle(savefile, overwrite) as hf:
            meta = hf.create_dataset("meta", shape=(0,))
            meta.attrs["class"] = self.__class__.__name__
            if (p := self.parameter_dimension) is not None:
                meta.attrs["parameter_dimension"] = p
            meta.attrs["nterms"] = self.nterms
            if self.entries is not None:
                group = hf.create_group("entries")
                for i, Ai in enumerate(self.entries):
                    name = f"A{i:d}"
                    if sparse.issparse(Ai):
                        utils.save_sparray(group.create_group(name), Ai)
                    else:
                        group.create_dataset(name, data=Ai)

    @classmethod
    def load(cls, loadfile: str, coeffs):
        """Load an affine parametric operator from an HDF5 file.

        Parameters
        ----------
        loadfile : str
            Path to the file where the operator was stored via :meth:`save()`.
        coeffs : iterable of callables
            Scalar-valued coefficient functions for each term of the affine
            expansion.
        Returns
        -------
        op : _AffineOperator
            Initialized operator object.
        """
        with utils.hdf5_loadhandle(loadfile) as hf:
            ClassName = hf["meta"].attrs["class"]
            if ClassName != cls.__name__:
                raise TypeError(
                    f"file '{loadfile}' contains '{ClassName}' "
                    f"object, use '{ClassName}.load()'"
                )
            nterms = int(hf["meta"].attrs["nterms"])

            entries = None
            if "entries" in hf:
                entries = []
                group = hf["entries"]
                for i in range(len(group)):
                    obj = group[f"A{i:d}"]
                    if isinstance(obj, h5py.Dataset):
                        entries.append(obj[:])
                    else:
                        entries.append(utils.load_sparray(obj))

            op = cls(coeffs, nterms=nterms, entries=entries, fromblock=False)

            if (key := "parameter_dimension") in hf["meta"].attrs:
                op.parameter_dimension = int(hf["meta"].attrs[key])
            return op


# Public affine operator classes ==============================================
class AffineConstantOperator(_AffineOperator):
    r"""Affine-parametric constant operator
    :math:`\Ophat_{\ell}(\qhat,\u;\bfmu)
    = \chat_{\ell}(\bfmu)
    = \sum_{a=0}^{A_{\ell}-1}\theta_\ell^{(a)}\!(\bfmu)\,\chat_{\ell}^{(a)}.`

    Here, each :math:`\theta_\ell^{(a)}:\RR^{p}\to\RR` is a scalar-valued
    function of the parameter vector
    and each :math:`\chat_{\ell}^{(a)} \in \RR^r` is a constant vector,
    see :class:`opinf.operators.ConstantOperator`.

    Parameters
    ----------
    coeffs : callable, (iterable of callables), or int
        Coefficient functions for the terms of the affine expansion.

        * If callable, it should receive a parameter vector
          :math:`\bfmu` and return the vector of affine coefficients,
          :math:`[~\theta_{\ell}^{(0)}(\bfmu)
          ~~\cdots~~\theta_{\ell}^{(A_{\ell}-1)}(\bfmu)~]\trp`.
          In this case, ``nterms`` is a required argument.
        * If an iterable, each entry should be a callable representing a
          single affine coefficient function :math:`\theta_{\ell}^{(a)}`.
        * If an integer :math:`p`, set :math:`A_{\ell} = p` and define
          :math:`\theta_{\ell}^{(i)}\!(\bfmu) = \mu_i`. This is equivalent to
          using ``coeffs=lambda mu: mu``, except the parameter dimension is
          also captured and ``nterms`` is not required.
    entries : list of ndarrays, or None
        Operator vectors for each term of the affine expansion, i.e.,
        :math:`\chat_{\ell}^{(0)},\ldots,\chat_{\ell}^{(A_{\ell}-1)}.`
        If not provided in the constructor, use :meth:`set_entries` later.
    fromblock : bool
        If ``True``, interpret ``entries`` as a horizontal concatenation
        of arrays; if ``False`` (default), interpret ``entries`` as a list
        of arrays.
    """

    _OperatorClass = ConstantOperator


class AffineLinearOperator(_AffineOperator):
    r"""Affine-parametric linear operator
    :math:`\Ophat_{\ell}(\qhat,\u;\bfmu)
    = \Ahat_{\ell}(\bfmu)\qhat = \left(
    \sum_{a=0}^{A_{\ell}-1}\theta_{\ell}^{(a)}\!(\bfmu)\,\Ahat_{\ell}^{(a)}
    \right)\qhat.`

    Here, each :math:`\theta_\ell^{(a)}:\RR^{p}\to\RR` is a scalar-valued
    function of the parameter vector
    and each :math:`\Ahat_{\ell}^{(a)} \in \RR^{r\times r}` is a constant
    matrix, see :class:`opinf.operators.LinearOperator`.

    Parameters
    ----------
    coeffs : callable, (iterable of callables), or int
        Coefficient functions for the terms of the affine expansion.

        * If callable, it should receive a parameter vector
          :math:`\bfmu` and return the vector of affine coefficients,
          :math:`[~\theta_{\ell}^{(0)}(\bfmu)
          ~~\cdots~~\theta_{\ell}^{(A_{\ell}-1)}(\bfmu)~]\trp`.
          In this case, ``nterms`` is a required argument.
        * If an iterable, each entry should be a callable representing a
          single affine coefficient function :math:`\theta_{\ell}^{(a)}`.
        * If an integer :math:`p`, set :math:`A_{\ell} = p` and define
          :math:`\theta_{\ell}^{(i)}\!(\bfmu) = \mu_i`. This is equivalent to
          using ``coeffs=lambda mu: mu``, except the parameter dimension is
          also captured and ``nterms`` is not required.
    entries : list of ndarrays, or None
        Operator matrices for each term of the affine expansion, i.e.,
        :math:`\Ahat_{\ell}^{(0)},\ldots,\Ahat_{\ell}^{(A_{\ell}-1)}.`
        If not provided in the constructor, use :meth:`set_entries` later.
    fromblock : bool
        If ``True``, interpret ``entries`` as a horizontal concatenation
        of arrays; if ``False`` (default), interpret ``entries`` as a list
        of arrays.
    """

    _OperatorClass = LinearOperator


class AffineQuadraticOperator(_AffineOperator):
    r"""Affine-parametric quadratic operator
    :math:`\Ophat_{\ell}(\qhat,\u;\bfmu)
    = \Hhat_{\ell}(\bfmu)[\qhat\otimes\qhat] = \left(
    \sum_{a=0}^{A_{\ell}-1}\theta_{\ell}^{(a)}\!(\bfmu)\,\Hhat_{\ell}^{(a)}
    \right)[\qhat\otimes\qhat].`

    Here, each :math:`\theta_\ell^{(a)}:\RR^{p}\to\RR` is a scalar-valued
    function of the parameter vector
    and each :math:`\Hhat_{\ell}^{(a)} \in \RR^{r\times r^2}` is a constant
    matrix, see :class:`opinf.operators.QuadraticOperator`.

    Parameters
    ----------
    coeffs : callable, (iterable of callables), or int
        Coefficient functions for the terms of the affine expansion.

        * If callable, it should receive a parameter vector
          :math:`\bfmu` and return the vector of affine coefficients,
          :math:`[~\theta_{\ell}^{(0)}(\bfmu)
          ~~\cdots~~\theta_{\ell}^{(A_{\ell}-1)}(\bfmu)~]\trp`.
          In this case, ``nterms`` is a required argument.
        * If an iterable, each entry should be a callable representing a
          single affine coefficient function :math:`\theta_{\ell}^{(a)}`.
        * If an integer :math:`p`, set :math:`A_{\ell} = p` and define
          :math:`\theta_{\ell}^{(i)}\!(\bfmu) = \mu_i`. This is equivalent to
          using ``coeffs=lambda mu: mu``, except the parameter dimension is
          also captured and ``nterms`` is not required.
    entries : list of ndarrays, or None
        Operator matrices for each term of the affine expansion, i.e.,
        :math:`\Hhat_{\ell}^{(0)},\ldots,\Hhat_{\ell}^{(A_{\ell}-1)}.`
        If not provided in the constructor, use :meth:`set_entries` later.
    fromblock : bool
        If ``True``, interpret ``entries`` as a horizontal concatenation
        of arrays; if ``False`` (default), interpret ``entries`` as a list
        of arrays.
    """

    _OperatorClass = QuadraticOperator


class AffineCubicOperator(_AffineOperator):
    r"""Affine-parametric cubic operator
    :math:`\Ophat_{\ell}(\qhat,\u;\bfmu)
    = \Ghat_{\ell}(\bfmu)[\qhat\otimes\qhat\otimes\qhat] = \left(
    \sum_{a=0}^{A_{\ell}-1}\theta_{\ell}^{(a)}\!(\bfmu)\,\Ghat_{\ell}^{(a)}
    \right)[\qhat\otimes\qhat\otimes\qhat].`

    Here, each :math:`\theta_\ell^{(a)}:\RR^{p}\to\RR` is a scalar-valued
    function of the parameter vector
    and each :math:`\Ghat_{\ell}^{(a)} \in \RR^{r\times r^3}` is a constant
    matrix, see :class:`opinf.operators.CubicOperator`.

    Parameters
    ----------
    coeffs : callable, (iterable of callables), or int
        Coefficient functions for the terms of the affine expansion.

        * If callable, it should receive a parameter vector
          :math:`\bfmu` and return the vector of affine coefficients,
          :math:`[~\theta_{\ell}^{(0)}(\bfmu)
          ~~\cdots~~\theta_{\ell}^{(A_{\ell}-1)}(\bfmu)~]\trp`.
          In this case, ``nterms`` is a required argument.
        * If an iterable, each entry should be a callable representing a
          single affine coefficient function :math:`\theta_{\ell}^{(a)}`.
        * If an integer :math:`p`, set :math:`A_{\ell} = p` and define
          :math:`\theta_{\ell}^{(i)}\!(\bfmu) = \mu_i`. This is equivalent to
          using ``coeffs=lambda mu: mu``, except the parameter dimension is
          also captured and ``nterms`` is not required.
    entries : list of ndarrays, or None
        Operator matrices for each term of the affine expansion, i.e.,
        :math:`\Ghat_{\ell}^{(0)},\ldots,\Ghat_{\ell}^{(A_{\ell}-1)}.`
        If not provided in the constructor, use :meth:`set_entries` later.
    fromblock : bool
        If ``True``, interpret ``entries`` as a horizontal concatenation
        of arrays; if ``False`` (default), interpret ``entries`` as a list
        of arrays.
    """

    _OperatorClass = CubicOperator


class AffineInputOperator(_AffineOperator, InputMixin):
    r"""Affine-parametric input operator
    :math:`\Ophat_{\ell}(\qhat,\u;\bfmu)
    = \Bhat_{\ell}(\bfmu)\u = \left(
    \sum_{a=0}^{A_{\ell}-1}\theta_{\ell}^{(a)}\!(\bfmu)\,\Bhat_{\ell}^{(a)}
    \right)\u.`

    Here, each :math:`\theta_\ell^{(a)}:\RR^{p}\to\RR` is a scalar-valued
    function of the parameter vector
    and each :math:`\Bhat_{\ell}^{(a)} \in \RR^{r\times m}` is a constant
    matrix, see :class:`opinf.operators.InputOperator`.

    Parameters
    ----------
    coeffs : callable, (iterable of callables), or int
        Coefficient functions for the terms of the affine expansion.

        * If callable, it should receive a parameter vector
          :math:`\bfmu` and return the vector of affine coefficients,
          :math:`[~\theta_{\ell}^{(0)}(\bfmu)
          ~~\cdots~~\theta_{\ell}^{(A_{\ell}-1)}(\bfmu)~]\trp`.
          In this case, ``nterms`` is a required argument.
        * If an iterable, each entry should be a callable representing a
          single affine coefficient function :math:`\theta_{\ell}^{(a)}`.
        * If an integer :math:`p`, set :math:`A_{\ell} = p` and define
          :math:`\theta_{\ell}^{(i)}\!(\bfmu) = \mu_i`. This is equivalent to
          using ``coeffs=lambda mu: mu``, except the parameter dimension is
          also captured and ``nterms`` is not required.
    entries : list of ndarrays, or None
        Operator matrices for each term of the affine expansion, i.e.,
        :math:`\Bhat_{\ell}^{(0)},\ldots,\Bhat_{\ell}^{(A_{\ell}-1)}.`
        If not provided in the constructor, use :meth:`set_entries` later.
    fromblock : bool
        If ``True``, interpret ``entries`` as a horizontal concatenation
        of arrays; if ``False`` (default), interpret ``entries`` as a list
        of arrays.
    """

    _OperatorClass = InputOperator

    @property
    def input_dimension(self):
        r"""Dimension of the input :math:`\u` that the operator acts on."""
        return None if self.entries is None else self.shape[1]


class AffineStateInputOperator(_AffineOperator, InputMixin):
    r"""Affine-parametric state-input operator
    :math:`\Ophat_{\ell}(\qhat,\u;\bfmu)
    = \Nhat_{\ell}(\bfmu)\qhat = \left(
    \sum_{a=0}^{A_{\ell}-1}\theta_{\ell}^{(a)}\!(\bfmu)\,\Nhat_{\ell}^{(a)}
    \right)[\u\otimes\qhat].`

    Here, each :math:`\theta_\ell^{(a)}:\RR^{p}\to\RR` is a scalar-valued
    function of the parameter vector
    and each :math:`\Nhat_{\ell}^{(a)} \in \RR^{r\times rm}` is a constant
    matrix, see :class:`opinf.operators.StateInputOperator`.

    Parameters
    ----------
    coeffs : callable, (iterable of callables), or int
        Coefficient functions for the terms of the affine expansion.

        * If callable, it should receive a parameter vector
          :math:`\bfmu` and return the vector of affine coefficients,
          :math:`[~\theta_{\ell}^{(0)}(\bfmu)
          ~~\cdots~~\theta_{\ell}^{(A_{\ell}-1)}(\bfmu)~]\trp`.
          In this case, ``nterms`` is a required argument.
        * If an iterable, each entry should be a callable representing a
          single affine coefficient function :math:`\theta_{\ell}^{(a)}`.
        * If an integer :math:`p`, set :math:`A_{\ell} = p` and define
          :math:`\theta_{\ell}^{(i)}\!(\bfmu) = \mu_i`. This is equivalent to
          using ``coeffs=lambda mu: mu``, except the parameter dimension is
          also captured and ``nterms`` is not required.
    entries : list of ndarrays, or None
        Operator matrices for each term of the affine expansion, i.e.,
        :math:`\Nhat_{\ell}^{(0)},\ldots,\Nhat_{\ell}^{(A_{\ell}-1)}.`
        If not provided in the constructor, use :meth:`set_entries` later.
    fromblock : bool
        If ``True``, interpret ``entries`` as a horizontal concatenation
        of arrays; if ``False`` (default), interpret ``entries`` as a list
        of arrays.
    """

    _OperatorClass = StateInputOperator

    @property
    def input_dimension(self):
        r"""Dimension of the input :math:`\u` that the operator acts on."""
        if self.entries is None:
            return None
        r, rm = self.shape
        return rm // r


# Utilities ===================================================================
def is_affine(obj) -> bool:
    """Return ``True`` if ``obj`` is a interpolated operator object."""
    return isinstance(obj, _AffineOperator)


def nonparametric_to_affine(OpClass: type) -> type:
    """Get the affine operator class corresponding to a nonparametric
    operator class.

    """
    for AffineClassName in __all__:
        AffineClass = eval(AffineClassName)
        if not isinstance(AffineClass, type) or not issubclass(
            AffineClass, _AffineOperator
        ):  # pragma: no cover
            continue
        if AffineClass._OperatorClass is OpClass:
            return AffineClass
    raise TypeError(
        f"_AffineOperator for class '{OpClass.__name__}' not found"
    )



# lift/_base.py
"""Template class for lifting transformations."""

__all__ = [
    "LifterTemplate",
]

import abc
import numpy as np
import scipy.linalg as la

from .. import errors, ddt, utils


class LifterTemplate(abc.ABC):
    """Template class for lifting transformations.

    Classes that inherit from this template must implement the methods
    :meth:`lift()` and :meth:`unlift()`. The optional :meth:`lift_ddts` method
    is used by the ROM class when snapshot time derivative data are available
    in the native state variables.

    See :class:`QuadraticLifter` for an example.
    """

    def __str__(self):
        """String representation: class name."""
        return self.__class__.__name__

    def __repr__(self):
        """Unique ID + string representation."""
        return utils.str2repr(self)

    @staticmethod
    @abc.abstractmethod
    def lift(states):  # pragma: no cover
        """Lift the native state variables to the learning variables.

        Parameters
        ----------
        states : (n, k) ndarray
            Native state variables.

        Returns
        -------
        lifted_states : (n_new, k) ndarray
            Learning variables.
        """
        raise NotImplementedError

    @staticmethod
    def lift_ddts(states, ddts):  # pragma: no cover
        """Lift the native state time derivatives to the time derivatives
        of the learning variables.

        Parameters
        ----------
        states : (n, k) ndarray
            Native state variables.
        ddts : (n, k) ndarray
            Time derivatives of the native state variables. Each column
            ``ddts[:, j]`` corresponds to the state vector ``states[:, j]``.

        Returns
        -------
        ddts : (n_new, k) ndarray
            Time derivatives of the learning variables.
        """
        return NotImplemented

    @staticmethod
    @abc.abstractmethod
    def unlift(lifted_states):  # pragma: no cover
        """Recover the native state variables from the learning variables.

        Parameters
        ----------
        lifted_states : (n, k) ndarray
            Learning variables.

        Returns
        -------
        states : (n, k) ndarray
            Native state variables.
        """
        raise NotImplementedError

    # Testing -----------------------------------------------------------------
    def verify(self, states, t=None, tol: float = 1e-4):
        r"""Verify that :meth:`lift` and :meth:`unlift` are consistent and that
        :meth:`lift_ddts`, if implemented, gives valid time derivatives.

        * The :meth:`lift` / :meth:`unlift` consistency check verifies that
          ``unlift(lift(states)) == states``.
        * The :meth:`lift_ddts` consistency check uses :meth:`opinf.ddt.ddt`
          to estimate the time derivatives of the states and the lifted
          states, then verfies that the relative difference between
          ``lift_ddts(states, opinf.ddt.ddt(states, t))`` and
          ``opinf.ddt.ddt(lift(states), t)`` is less than ``tol``.
          If this check fails, consider using a finer time mesh.

        Parameters
        ----------
        states : (n, k) ndarray
            Native state variables.
        t : (k,) ndarray or None
            Time domain corresponding to the states.
            Only required if :meth:`lift_ddts` is implemented.
        tol : float > 0
            Tolerance for the finite difference check of :meth:`lift_ddts`.
            Only used if :meth:`lift_ddts` is implemented.
        """
        # Verify lift() and unlift() are inverses.
        lifted_states = self.lift(states)
        if (k1 := lifted_states.shape[1]) != states.shape[1]:
            raise errors.VerificationError(
                f"{k1} = lift(states).shape[1] "
                f"!= states.shape[1] = {states.shape[1]}"
            )

        unlifted_states = self.unlift(lifted_states)
        if (shape := unlifted_states.shape) != states.shape:
            raise errors.VerificationError(
                f"{shape} = unlift(lift(states)).shape "
                f"!= states.shape = {states.shape}"
            )
        if not np.allclose(unlifted_states, states):
            raise errors.VerificationError("unlift(lift(states)) != states")
        print("lift() and unlift() are consistent")

        # Finite difference checks for lift_ddts().
        if self.lift_ddts(states, states) is NotImplemented:
            return
        if t is None:
            raise ValueError(
                "time domain 't' required for finite difference check"
            )
        lifted_ddts = self.lift_ddts(states, ddt.ddt(states, t))
        if (shape := lifted_ddts.shape) != (shape2 := lifted_states.shape):
            raise errors.VerificationError(
                f"{shape} = lift_ddts(states, ddts).shape "
                f"!= lift(states).shape = {shape2}"
            )
        lddts_est = ddt.ddt(lifted_states, t)
        if (
            diff := la.norm(lifted_ddts - lddts_est) / la.norm(lddts_est)
        ) > tol:
            raise errors.VerificationError(
                "lift_ddts() failed finite difference check,\n\t"
                "|| lift_ddts(states, d/dt[states]) - d/dt[lift(states)] || "
                f" / || d/dt[lift(states)] || = {diff} > tol = {tol}"
            )
        print("lift() and lift_ddts() are consistent")
